from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Tuple

from lean_dojo_v2.lean_dojo.data_extraction.lean import Pos


def parse_pos(pos_str):
    """
    Parses a position string or list into a Pos object.

    Args:
        pos_str (str or list): The position data, either as a string in the format 'Pos(x, y)'
                               or as a list [x, y].

    Returns:
        Pos: A Pos object initialized with the parsed coordinates.

    Raises:
        ValueError: If the input format is neither a string nor a list.
    """
    if isinstance(pos_str, str):
        # pos_str came from a JSON file
        pos_parts = (
            pos_str.replace("Pos", "").replace("(", "").replace(")", "").split(",")
        )
        return Pos(int(pos_parts[0]), int(pos_parts[1]))
    elif isinstance(pos_str, list):
        # pos_str came from a dictionary initialization
        return Pos(*pos_str)
    else:
        raise ValueError(f"Unexpected format for Pos: {pos_str}")


@dataclass
class Annotation:
    """
    Annotation class represents a code annotation with its full name, definition path,
    and position details.

    Attributes:
        full_name (str): The full name of the annotation.
        def_path (str): The file path where the annotation is defined.
        def_pos (Pos): The starting position of the annotation definition.
        def_end_pos (Pos): The ending position of the annotation definition.
    """

    full_name: str
    def_path: str
    def_pos: Pos
    def_end_pos: Pos

    @classmethod
    def from_dict(cls, data: Dict) -> Annotation:
        if not all(
            key in data for key in ["full_name", "def_path", "def_pos", "def_end_pos"]
        ):
            raise ValueError("Invalid Annotation data format")
        return cls(
            full_name=data["full_name"],
            def_path=data["def_path"],
            def_pos=parse_pos(data["def_pos"]),
            def_end_pos=parse_pos(data["def_end_pos"]),
        )

    def to_dict(self) -> Dict:
        return {
            "full_name": self.full_name,
            "def_path": self.def_path,
            "def_pos": repr(self.def_pos),
            "def_end_pos": repr(self.def_end_pos),
        }


@dataclass
class AnnotatedTactic:
    """
    AnnotatedTactic is a data class that represents a tactic with its annotations
    and states before and after its application.

    Attributes:
        tactic (str): The tactic applied.
        annotated_tactic (Tuple[str, List[Annotation]]): A tuple containing the tactic and a list of annotations.
        state_before (str): The state before the tactic is applied.
        state_after (str): The state after the tactic is applied.
    """

    tactic: str
    annotated_tactic: Tuple[str, List[Annotation]]
    state_before: str
    state_after: str

    @classmethod
    def from_dict(cls, data: Dict) -> AnnotatedTactic:
        if not all(
            key in data
            for key in ["tactic", "annotated_tactic", "state_before", "state_after"]
        ):
            raise ValueError("Invalid AnnotatedTactic data format")
        return cls(
            tactic=data["tactic"],
            annotated_tactic=(
                data["annotated_tactic"][0],
                [Annotation.from_dict(a) for a in data["annotated_tactic"][1]],
            ),
            state_before=data["state_before"],
            state_after=data["state_after"],
        )

    def to_dict(self) -> Dict:
        return {
            "tactic": self.tactic,
            "annotated_tactic": [
                self.annotated_tactic[0],
                [a.to_dict() for a in self.annotated_tactic[1]],
            ],
            "state_before": self.state_before,
            "state_after": self.state_after,
        }



"""
Utility functions for Lean version management and compatibility checking.
"""

import re
from pathlib import Path

from lean_dojo_v2.utils.constants import LEAN4_BUILD_DIR, LEAN4_PACKAGES_DIR

# Regex pattern for parsing Lean 4 toolchain versions
_LEAN4_VERSION_REGEX = re.compile(
    r"leanprover/lean4:(?P<version>v\d+\.\d+\.\d+(?:-rc\d+)?)"
)


def get_lean4_version_from_config(toolchain: str) -> str:
    """Return the required Lean version given a ``lean-toolchain`` config."""
    m = _LEAN4_VERSION_REGEX.fullmatch(toolchain.strip())
    assert m is not None, "Invalid config."
    return m["version"]


def is_supported_version(v: str) -> bool:
    """
    Check if ``v`` is at least `v4.3.0` and at most `v4.30.0`.
    Note: Lean versions are generally not backwards-compatible. Also, the Lean FRO
    keeps bumping the default versions of repos to the latest version, which is
    not necessarily the latest stable version. So, we need to be careful about
    what we choose to support.
    """
    max_version = 30
    min_version = 3
    if not v.startswith("v"):
        return False
    v = v[1:]
    major, minor, patch = [int(_) for _ in v.split("-")[0].split(".")]
    if (
        major < 4
        or (major == 4 and minor < min_version)
        or (major == 4 and minor > max_version)
        or (major == 4 and minor == max_version and patch > 1)
    ):
        return False
    if (
        major > 4
        or (major == 4 and minor > min_version)
        or (major == 4 and minor == min_version and patch > 0)
    ):
        return True


def _from_lean_path(root_dir: Path, path: Path, repo, ext: str) -> Path:
    assert path.suffix == ".lean"
    if path.is_absolute():
        path = path.relative_to(root_dir)

    assert root_dir.name != "lean4"
    if path.is_relative_to(LEAN4_PACKAGES_DIR / "lean4/src/lean/lake"):
        # E.g., "lake-packages/lean4/src/lean/lake/Lake/CLI/Error.lean"
        p = path.relative_to(LEAN4_PACKAGES_DIR / "lean4/src/lean/lake")
        return LEAN4_PACKAGES_DIR / "lean4/lib/lean" / p.with_suffix(ext)
    elif path.is_relative_to(LEAN4_PACKAGES_DIR / "lean4/src"):
        # E.g., "lake-packages/lean4/src/lean/Init.lean"
        p = path.relative_to(LEAN4_PACKAGES_DIR / "lean4/src").with_suffix(ext)
        return LEAN4_PACKAGES_DIR / "lean4/lib" / p
    elif path.is_relative_to(LEAN4_PACKAGES_DIR):
        # E.g., "lake-packages/std/Std.lean"
        p = path.relative_to(LEAN4_PACKAGES_DIR).with_suffix(ext)
        repo_name = p.parts[0]
        return (
            LEAN4_PACKAGES_DIR
            / repo_name
            / LEAN4_BUILD_DIR
            / "ir"
            / p.relative_to(repo_name)
        )
    else:
        # E.g., "Mathlib/LinearAlgebra/Basics.lean"
        return LEAN4_BUILD_DIR / "ir" / path.with_suffix(ext)


def to_xml_path(root_dir: Path, path: Path, repo) -> Path:
    return _from_lean_path(root_dir, path, repo, ext=".trace.xml")


def to_dep_path(root_dir: Path, path: Path, repo) -> Path:
    return _from_lean_path(root_dir, path, repo, ext=".dep_paths")


def to_json_path(root_dir: Path, path: Path, repo) -> Path:
    return _from_lean_path(root_dir, path, repo, ext=".ast.json")


def to_lean_path(root_dir: Path, path: Path, repo) -> Path:
    if path.is_absolute():
        path = path.relative_to(root_dir)

    if path.suffix in (".xml", ".json"):
        path = path.with_suffix("").with_suffix(".lean")
    else:
        assert path.suffix == ".dep_paths"
        path = path.with_suffix(".lean")

    assert root_dir.name != "lean4"
    if path == LEAN4_PACKAGES_DIR / "lean4/lib/lean/Lake.lean":
        return LEAN4_PACKAGES_DIR / "lean4/src/lean/lake/Lake.lean"
    elif path == LEAN4_PACKAGES_DIR / "lean4/lib/lean/LakeMain.lean":
        return LEAN4_PACKAGES_DIR / "lean4/src/lean/lake/LakeMain.lean"
    elif path.is_relative_to(LEAN4_PACKAGES_DIR / "lean4/lib/lean/Lake"):
        # E.g., "lake-packages/lean4/lib/lean/Lake/Util/List.lean"
        p = path.relative_to(LEAN4_PACKAGES_DIR / "lean4/lib/lean/Lake")
        return LEAN4_PACKAGES_DIR / "lean4/src/lean/lake/Lake" / p
    elif path.is_relative_to(LEAN4_PACKAGES_DIR / "lean4/lib"):
        # E.g., "lake-packages/lean4/lib/lean/Init.lean"
        p = path.relative_to(LEAN4_PACKAGES_DIR / "lean4/lib")
        return LEAN4_PACKAGES_DIR / "lean4/src" / p
    elif path.is_relative_to(LEAN4_PACKAGES_DIR):
        # E.g., "lake-packages/std/build/ir/Std.lean"
        p = path.relative_to(LEAN4_PACKAGES_DIR)
        repo_name = p.parts[0]
        return (
            LEAN4_PACKAGES_DIR
            / repo_name
            / p.relative_to(Path(repo_name) / LEAN4_BUILD_DIR / "ir")
        )
    else:
        # E.g., ".lake/build/ir/Mathlib/LinearAlgebra/Basics.lean" or "build/ir/Mathlib/LinearAlgebra/Basics.lean"
        assert path.is_relative_to(LEAN4_BUILD_DIR / "ir"), path
        return path.relative_to(LEAN4_BUILD_DIR / "ir")
    assert major == 4 and minor == min_version and patch == 0
    return True



from lean_dojo_v2.prover.external_prover import ExternalProver

from .base_agent import BaseAgent


class ExternalAgent(BaseAgent):
    def __init__(
        self,
        model_name: str = "deepseek-ai/DeepSeek-Prover-V2-671B:novita",
    ):
        super().__init__()
        self.model_name = model_name

    def _get_build_deps(self) -> bool:
        """ExternalAgent doesn't build dependencies by default."""
        return False

    def _setup_prover(self):
        """Set up the RetrievalProver for ExternalAgent."""
        self.prover = ExternalProver(model_name=self.model_name)


def main():
    """
    Main function to run LeanAgent.
    """
    url = "https://github.com/durant42040/lean4-example"
    commit = "005de00d03f1aaa32cb2923d5e3cbaf0b954a192"

    agent = ExternalAgent(model_name="deepseek-ai/DeepSeek-Prover-V2-671B:novita")
    agent.setup_github_repository(url=url, commit=commit)
    agent.prove()


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List

from lean_dojo_v2.lean_dojo.data_extraction.lean import Pos

from .annotations import parse_pos


@dataclass
class Premise:
    """
    A class representing a Premise with various attributes.

    Attributes:
        full_name (str): The full name of the premise.
        code (str): The code associated with the premise.
        start (Pos): The starting position of the premise.
        end (Pos): The ending position of the premise.
        kind (str): The kind or type of the premise.
    """

    full_name: str
    code: str
    start: Pos
    end: Pos
    kind: str

    @classmethod
    def from_dict(cls, data: Dict) -> Premise:
        if not all(
            key in data for key in ["full_name", "code", "start", "end", "kind"]
        ):
            raise ValueError("Invalid Premise data format")
        return cls(
            full_name=data["full_name"],
            code=data["code"],
            start=parse_pos(data["start"]),
            end=parse_pos(data["end"]),
            kind=data["kind"],
        )

    def to_dict(self) -> Dict:
        return {
            "full_name": self.full_name,
            "code": self.code,
            "start": repr(self.start),
            "end": repr(self.end),
            "kind": self.kind,
        }


@dataclass
class PremiseFile:
    """
    Represents a file containing premises and their associated imports.

    Attributes:
        path (Path): The file path.
        imports (List[str]): A list of import statements.
        premises (List[Premise]): A list of premises.
    """

    path: Path
    imports: List[str]
    premises: List[Premise]

    @classmethod
    def from_dict(cls, data: Dict) -> PremiseFile:
        if not all(key in data for key in ["path", "imports", "premises"]):
            raise ValueError("Invalid PremiseFile data format")
        return cls(
            path=Path(data["path"]),
            imports=data["imports"],
            premises=[Premise.from_dict(p) for p in data["premises"]],
        )

    def to_dict(self) -> Dict:
        return {
            "path": str(self.path),
            "imports": self.imports,
            "premises": [p.to_dict() for p in self.premises],
        }



"""This module define classes for repos, files, and theorems in Lean.
Objects of these classes contain only surface information, without extracting any trace.
"""

import json
import os
import re
import shutil
import tempfile
import time
import urllib
import webbrowser
from dataclasses import dataclass, field
from enum import Enum
from functools import cache
from pathlib import Path
from typing import Any, Dict, Generator, Iterator, List, Optional, Tuple, Union

import toml
from git import BadName, Repo
from github import Auth, Github
from github.GithubException import GithubException
from github.Repository import Repository
from loguru import logger

from lean_dojo_v2.utils.common import read_url, url_exists
from lean_dojo_v2.utils.constants import LEAN4_URL, TMP_DIR
from lean_dojo_v2.utils.filesystem import is_git_repo, working_directory

from .cache import cache as repo_cache

GITHUB_ACCESS_TOKEN = os.getenv("GITHUB_ACCESS_TOKEN", None)
"""GiHub personal access token is optional. 
If provided, it can increase the rate limit for GitHub API calls.
"""

if GITHUB_ACCESS_TOKEN:
    logger.debug("Using GitHub personal access token for authentication")
    GITHUB = Github(auth=Auth.Token(GITHUB_ACCESS_TOKEN))
    GITHUB.get_user().login
else:
    logger.debug(
        "Using GitHub without authentication. Don't be surprised if you hit the API rate limit."
    )
    GITHUB = Github()

LEAN4_REPO = None
"""The GitHub Repo for Lean 4 itself."""

LEAN4_NIGHTLY_REPO = None
"""The GitHub Repo for Lean 4 nightly releases."""

_URL_REGEX = re.compile(r"(?P<url>.*?)/*")

_SSH_TO_HTTPS_REGEX = re.compile(r"git@github\.com:(?P<user>.+)/(?P<repo>.+?)(\.git)?")

REPO_CACHE_PREFIX = "repos"


class RepoType(Enum):
    GITHUB = 0
    REMOTE = 1  # Remote but not GitHub.
    LOCAL = 2


def normalize_url(url: str, repo_type: RepoType = RepoType.GITHUB) -> str:
    if repo_type == RepoType.LOCAL:  # Convert to absolute path if local.
        return os.path.abspath(url)
    # Remove trailing `/`.
    url = _URL_REGEX.fullmatch(url)["url"]  # type: ignore
    return ssh_to_https(url)


def ssh_to_https(url: str) -> str:
    m = _SSH_TO_HTTPS_REGEX.fullmatch(url)
    return f"https://github.com/{m.group('user')}/{m.group('repo')}" if m else url


def get_repo_type(url: str) -> Optional[RepoType]:
    """Get the type of the repository.

    Args:
        url (str): The URL of the repository.
    Returns:
        Optional[str]: The type of the repository (None if the repo cannot be found).
    """
    url = ssh_to_https(url)
    parsed_url = urllib.parse.urlparse(url)  # type: ignore
    if parsed_url.scheme in ["http", "https"]:
        # Case 1 - GitHub URL.
        if "github.com" in url:
            if not url.startswith("https://"):
                logger.warning(f"{url} should start with https://")
                return None
            else:
                return RepoType.GITHUB
        # Case 2 - remote URL.
        elif url_exists(url):  # Not check whether it is a git URL
            return RepoType.REMOTE
    # Case 3 - local path
    elif is_git_repo(Path(parsed_url.path)):
        return RepoType.LOCAL
    logger.warning(f"{url} is not a valid URL")
    return None


def _split_git_url(url: str) -> Tuple[str, str]:
    """Split a Git URL into user name and repo name."""
    if url.endswith("/"):
        url = url[:-1]
        assert not url.endswith("/"), f"Unexpected URL: {url}"
    fields = url.split("/")
    user_name = fields[-2]
    repo_name = fields[-1]
    return user_name, repo_name


def _format_cache_dirname(url: str, commit: str) -> str:
    user_name, repo_name = _split_git_url(url)
    repo_type = get_repo_type(url)
    assert repo_type is not None, f"Invalid url {url}"
    if repo_type == RepoType.GITHUB:
        return f"{user_name}-{repo_name}-{commit}"
    else:  # git repo
        return f"gitpython-{repo_name}-{commit}"


@cache
def url_to_repo(
    url: str,
    num_retries: int = 2,
    repo_type: Optional[RepoType] = None,
    tmp_dir: Optional[Path] = None,
) -> Union[Repo, Repository]:
    """Convert a URL to a Repo object.

    Args:
        url (str): The URL of the repository.
        num_retries (int): Number of retries in case of failure.
        repo_type (Optional[RepoType]): The type of the repository. Defaults to None.
        tmp_dir (Optional[Path]): The temporary directory to clone the repo to. Defaults to None.

    Returns:
        Repo: A Git Repo object.
    """
    url = normalize_url(url)
    backoff = 1
    if tmp_dir is None:
        tmp_dir = (TMP_DIR or Path("/tmp")) / next(tempfile._get_candidate_names())  # type: ignore
    repo_type = repo_type or get_repo_type(url)
    assert repo_type is not None, f"Invalid url {url}"
    if repo_type == RepoType.GITHUB:
        return GITHUB.get_repo("/".join(url.split("/")[-2:]))
    with working_directory(tmp_dir):
        repo_name = os.path.basename(url)
        if repo_type == RepoType.LOCAL:
            assert is_git_repo(url), f"Local path {url} is not a git repo"
            shutil.copytree(url, repo_name)
            return Repo(repo_name)
        else:
            return Repo.clone_from(url, repo_name)


@cache
def get_latest_commit(url: str) -> str:
    """Get the hash of the latest commit of the Git repo at ``url``."""
    repo = url_to_repo(url)
    if isinstance(repo, Repository):
        return repo.get_branch(repo.default_branch).commit.sha
    else:
        return repo.head.commit.hexsha


def cleanse_string(s: Union[str, Path]) -> str:
    """Replace : and / with _ in a string."""
    return str(s).replace("/", "_").replace(":", "_")


@cache
def _to_commit_hash(repo: Union[Repository, Repo], label: str) -> str:
    """Convert a tag or branch to a commit hash."""
    if isinstance(repo, Repository):  # GitHub repository
        try:
            return repo.get_commit(label).sha
        except GithubException as ex:
            raise ValueError(f"Invalid tag or branch: `{label}` for {repo.name}")
    else:  # Local or remote Git repository
        assert isinstance(repo, Repo)
        logger.debug(
            f"Querying the commit hash for {repo.working_dir} repository {label}"
        )
        try:
            # Resolve the label to a commit hash
            return repo.commit(label).hexsha
        except Exception as ex:
            raise ValueError(f"Error converting ref to commit hash: {ex}")


@dataclass(eq=True, unsafe_hash=True)
class Pos:
    """Position in source files.

    We use 1-index to keep it consistent with code editors such as Visual Studio Code.
    """

    line_nb: int
    """Line number
    """

    column_nb: int
    """Column number
    """

    @classmethod
    def from_str(cls, s: str) -> "Pos":
        """Construct a :class:`Pos` object from its string representation, e.g., :code:`"(323, 1109)"`."""
        assert s.startswith("(") and s.endswith(
            ")"
        ), f"Invalid string representation of a position: {s}"
        line, column = s[1:-1].split(",")
        line_nb = int(line)
        column_nb = int(column)
        return cls(line_nb, column_nb)

    def __iter__(self) -> Generator[int, None, None]:
        yield self.line_nb
        yield self.column_nb

    def __repr__(self) -> str:
        return repr(tuple(self))

    def __lt__(self, other):
        return self.line_nb < other.line_nb or (
            self.line_nb == other.line_nb and self.column_nb < other.column_nb
        )

    def __le__(self, other):
        return self < other or self == other


@dataclass(frozen=True)
class LeanFile:
    """A Lean source file (:file:`*.lean`)."""

    root_dir: Path = field(repr=False)
    """Root directory of the traced repo this :class:`LeanFile` object belongs to.

    ``root_dir`` must be an absolute path, e.g., :file:`/home/kaiyu/traced_lean-example/lean-example`
    """

    path: Path
    """Relative path w.r.t. ``root_dir``
    
    E.g., :file:`lean-example/src/example.lean`
    """

    code: List[str] = field(init=False, repr=False)
    """Raw source code as a list of lines."""

    endwith_newline: bool = field(init=False, repr=False)
    """Whether the last line ends with a newline."""

    num_bytes: List[int] = field(init=False, repr=False)
    """The number of UTF-8 bytes of each line, including newlines.
    """

    def __post_init__(self) -> None:
        assert (
            self.root_dir.is_absolute()
        ), f"Root directory must be an absolute path: {self.root_dir}"
        assert self.path.suffix == ".lean", f"File extension must be .lean: {self.path}"
        assert not self.path.is_absolute(), f"Path must be a relative path: {self.path}"

        code = []
        endwith_newline = None
        num_bytes = []

        for line in self.abs_path.open("rb"):
            if b"\r\n" in line:
                raise RuntimeError(
                    f"{self.abs_path} contains Windows-style line endings. This is discouraged (see https://github.com/leanprover-community/mathlib4/pull/6506)."
                )
            if line.endswith(b"\n"):
                endwith_newline = True
                line = line[:-1]
            else:
                endwith_newline = False
            code.append(line.decode("utf-8"))
            num_bytes.append(len(line) + 1)

        object.__setattr__(self, "code", code)
        object.__setattr__(self, "endwith_newline", endwith_newline)
        object.__setattr__(self, "num_bytes", num_bytes)

    @property
    def abs_path(self) -> Path:
        """Absolute path of a :class:`LeanFile` object.

        E.g., :file:`/home/kaiyu/traced_lean-example/lean-example/src/example.lean`
        """
        return self.root_dir / self.path

    @property
    def num_lines(self) -> int:
        """Number of lines in a source file."""
        return len(self.code)

    def num_columns(self, line_nb: int) -> int:
        """Number of columns in a source file."""
        return len(self.get_line(line_nb))

    @property
    def start_pos(self) -> Pos:
        """Return the start position of a source file.

        Returns:
            Pos: A :class:`Pos` object representing the start of this file.
        """
        return Pos(1, 1)

    @property
    def end_pos(self) -> Pos:
        """Return the end position of a source file.

        Returns:
            Pos: A :class:`Pos` object representing the end of this file.
        """
        # Line and column numbers are 1-indexed by default.
        if self.is_empty():
            return self.start_pos
        line_nb = self.num_lines
        column_nb = 1 + len(self.code[-1])
        return Pos(line_nb, column_nb)

    def is_empty(self) -> bool:
        return len(self.code) == 0

    def position_to_pos(self, pos_val: Any) -> Pos:
        """Convert a position value (int or dict) from Lean 4 JSON to a :class:`Pos` object.

        Handles both legacy (integer byte index) and newer (dict with byteIdx) formats.
        """
        if isinstance(pos_val, int):
            return self.convert_pos(pos_val)
        if isinstance(pos_val, dict):
            byte_idx = pos_val.get("byteIdx")
            if byte_idx is not None:
                return self.convert_pos(byte_idx)
            raw = pos_val.get("raw")
            if isinstance(raw, dict) and raw.get("byteIdx") is not None:
                return self.convert_pos(raw["byteIdx"])
            if "1" in pos_val and isinstance(pos_val["1"], int):
                return self.convert_pos(pos_val["1"])
            line = pos_val.get("line")
            col = pos_val.get("column") or pos_val.get("character")
            if line is not None and col is not None:
                return Pos(line + 1, col + 1)
        raise ValueError(
            f"Cannot parse position: expected int or dict with byteIdx/line+column, got {type(pos_val).__name__}"
        )

    def convert_pos(self, byte_idx: int) -> Pos:
        """Convert a byte index (:code:`String.Pos` in Lean 4) to a :class:`Pos` object."""
        n = 0
        for i, num_bytes in enumerate(self.num_bytes, start=1):
            n += num_bytes
            if n == byte_idx and i == self.num_lines:
                byte_idx -= 1
            if n > byte_idx:
                line_byte_idx = byte_idx - (n - num_bytes)
                if line_byte_idx == 0:
                    return Pos(i, 1)

                line = self.get_line(i)
                m = 0

                for j, c in enumerate(line, start=1):
                    m += len(c.encode("utf-8"))
                    if m >= line_byte_idx:
                        return Pos(i, j + 1)

        raise ValueError(f"Invalid byte index {byte_idx} in {self.path}.")

    def offset(self, pos: Pos, delta: int) -> Pos:
        """Off set a position by a given number."""
        line_nb, column_nb = pos
        num_columns = len(self.get_line(line_nb)) - column_nb + 1
        if delta <= num_columns:
            return Pos(line_nb, column_nb + delta)
        delta_left = delta - num_columns - 1

        for i in range(line_nb, self.num_lines):
            line = self.code[i]
            l = len(line)
            if delta_left <= l:
                return Pos(i + 1, delta_left + 1)
            delta_left -= l + 1

        if delta_left == 0 and self.endwith_newline:
            return Pos(self.num_lines + 1, 1)

        raise ValueError(f"Invalid offset {delta} in {self.path}: {pos}.")

    def get_line(self, line_nb: int) -> str:
        """Return a given line of the source file.

        Args:
            line_nb (int): Line number (1-indexed).
        """
        return self.code[line_nb - 1]

    def __getitem__(self, key) -> str:
        """Return a code segment given its start/end positions.

        This enables ``lean_file[start:end]``.

        Args:
            key (slice): A slice of two :class:`Pos` objects for the start/end of the code segment.
        """
        assert isinstance(key, slice) and key.step is None
        if key.start is None:
            start_line = start_column = 1
        else:
            start_line, start_column = key.start
        if key.stop is None:
            end_line = self.num_lines
            end_column = 1 + len(self.get_line(end_line))
        else:
            end_line, end_column = key.stop
        if start_line == end_line:
            assert start_column <= end_column
            return self.get_line(start_line)[start_column - 1 : end_column - 1]
        else:
            assert start_line < end_line
            code_slice = [self.code[start_line - 1][start_column - 1 :]]
            for line_nb in range(start_line + 1, end_line):
                code_slice.append(self.get_line(line_nb))
            code_slice.append(self.get_line(end_line)[: end_column - 1])
            return "\n".join(code_slice)


_COMMIT_REGEX = re.compile(r"[0-9a-z]+")
_LEAN4_VERSION_REGEX = re.compile(r"leanprover/lean4:(?P<version>.+?)")


def is_commit_hash(s: str):
    """Check if a string is a valid commit hash."""
    return len(s) == 40 and _COMMIT_REGEX.fullmatch(s)


def get_lean4_version_from_config(toolchain: str) -> str:
    """Return the required Lean version given a ``lean-toolchain`` config."""
    m = _LEAN4_VERSION_REGEX.fullmatch(toolchain.strip())
    assert m is not None, "Invalid config."
    v = m["version"]
    if not v.startswith("v") and v[0].isnumeric():
        v = "v" + v
    return v


def get_lean4_commit_from_config(config_dict: Dict[str, Any]) -> str:
    """Return the required Lean commit given a ``lean-toolchain`` config."""
    global LEAN4_REPO
    if LEAN4_REPO is None:
        LEAN4_REPO = GITHUB.get_repo("leanprover/lean4")
    assert "content" in config_dict, "config_dict must have a 'content' field"
    version = get_lean4_version_from_config(config_dict["content"].strip())
    if version.startswith("nightly-"):
        global LEAN4_NIGHTLY_REPO
        if LEAN4_NIGHTLY_REPO is None:
            LEAN4_NIGHTLY_REPO = GITHUB.get_repo("leanprover/lean4-nightly")
        return _to_commit_hash(LEAN4_NIGHTLY_REPO, version)
    else:
        return _to_commit_hash(LEAN4_REPO, version)


URL = str
TAG = str
COMMIT = str


@dataclass(frozen=True)
class RepoInfoCache:
    """To minize the number of network requests, we cache and re-use the info
    of all repos, assuming it does not change during the execution of LeanDojo."""

    tag2commit: Dict[Tuple[URL, TAG], COMMIT] = field(default_factory=dict)
    lean_version: Dict[Tuple[URL, COMMIT], str] = field(default_factory=dict)


info_cache = RepoInfoCache()


_LAKEFILE_LEAN_GIT_REQUIREMENT_REGEX = re.compile(
    r"require\s+(?P<name>\S+)\s+from\s+git\s+\"(?P<url>.+?)\"(\s+@\s+\"(?P<rev>\S+)\")?"
)

_LAKEFILE_LEAN_LOCAL_REQUIREMENT_REGEX = re.compile(r"require \S+ from \"")

_LAKEFILE_TOML_REQUIREMENT_REGEX = re.compile(r"(?<=\[\[require\]\]).+(?=\n\n)")


def is_supported_version(v) -> bool:
    """Check if ``v`` is at least `v4.3.0-rc2`."""
    if not v.startswith("v"):
        return False
    v = v[1:]
    major, minor, patch = [int(_) for _ in v.split("-")[0].split(".")]
    if major < 4 or (major == 4 and minor < 3):
        return False
    if (
        major > 4
        or (major == 4 and minor > 3)
        or (major == 4 and minor == 3 and patch > 0)
    ):
        return True
    assert major == 4 and minor == 3 and patch == 0
    if "4.3.0-rc" in v:
        rc = int(v.split("-")[1][2:])
        return rc >= 2
    else:
        return True


@dataclass(frozen=True)
class LeanGitRepo:
    """Git repo of a Lean project."""

    url: str
    """The repo's URL.

    It can be a GitHub URL that starts with https:// or git@github.com, a local path, or any other valid Git URL.
    """

    commit: str
    """The repo's commit hash.
    
    You can also use tags such as ``v3.5.0``. They will be converted to commit hashes.
    """

    repo: Union[Repository, Repo] = field(init=False, repr=False)
    """A :class:`github.Repository` object for GitHub repos or
    a :class:`git.Repo` object for local or remote Git repos.
    """

    lean_version: str = field(init=False, repr=False)
    """Required Lean version.
    """

    repo_type: RepoType = field(init=False, repr=False)
    """Type of the repo. It can be ``GITHUB``, ``LOCAL`` or ``REMOTE``.
    """

    def __post_init__(self) -> None:
        repo_type = get_repo_type(self.url)
        if repo_type is None:
            raise ValueError(f"{self.url} is not a valid URL")
        object.__setattr__(self, "repo_type", repo_type)
        object.__setattr__(self, "url", normalize_url(self.url, repo_type=repo_type))
        # set repo and commit
        if repo_type == RepoType.GITHUB:
            repo = url_to_repo(self.url, repo_type=repo_type)
        else:
            # get repo from cache
            rel_cache_dir = lambda url, commit: Path(
                f"{REPO_CACHE_PREFIX}/{_format_cache_dirname(url, commit)}/{self.name}"
            )
            cache_repo_dir = repo_cache.get(rel_cache_dir(self.url, self.commit))
            if cache_repo_dir is None:
                with working_directory() as tmp_dir:
                    repo = url_to_repo(self.url, repo_type=repo_type, tmp_dir=tmp_dir)
                    commit = _to_commit_hash(repo, self.commit)
                    cache_repo_dir = repo_cache.store(
                        repo.working_dir, rel_cache_dir(self.url, commit)
                    )
            repo = Repo(cache_repo_dir)
        # Convert tags or branches to commit hashes
        if not is_commit_hash(self.commit):
            if (self.url, self.commit) in info_cache.tag2commit:
                commit = info_cache.tag2commit[(self.url, self.commit)]
            else:
                commit = _to_commit_hash(repo, self.commit)
                assert is_commit_hash(commit), f"Invalid commit hash: {commit}"
                info_cache.tag2commit[(self.url, commit)] = commit
            object.__setattr__(self, "commit", commit)
        object.__setattr__(self, "repo", repo)

        # Determine the required Lean version.
        if (self.url, self.commit) in info_cache.lean_version:
            lean_version = info_cache.lean_version[(self.url, self.commit)]
        if self.is_lean4:
            lean_version = "latest"  # lean4 itself
        else:
            config = self.get_config("lean-toolchain")
            lean_version = get_lean4_version_from_config(config["content"])
            if not is_supported_version(lean_version):
                logger.warning(
                    f"{self} relies on an unsupported Lean version: {lean_version}"
                )
        info_cache.lean_version[(self.url, self.commit)] = lean_version
        object.__setattr__(self, "lean_version", lean_version)

    @classmethod
    def from_path(cls, path: Union[Path, str]) -> "LeanGitRepo":
        """Construct a :class:`LeanGitRepo` object from the path to a local Git repo."""
        commit = Repo(path).head.commit.hexsha
        return cls(str(path), commit)

    @property
    def name(self) -> str:
        return os.path.basename(self.url)

    @property
    def is_lean4(self) -> bool:
        return self.url == LEAN4_URL

    @property
    def commit_url(self) -> str:
        return f"{self.url}/tree/{self.commit}"

    def get_cache_dirname(self) -> Path:
        """Return the formatted cache directory name"""
        assert is_commit_hash(self.commit), f"Invalid commit hash: {self.commit}"
        return Path(_format_cache_dirname(self.url, self.commit))

    def __str__(self) -> str:
        """Directory-friendly string representation.

        Example: "repo-name_commit-hash"
        """
        return f"{self.name}_{self.commit}"

    def show(self) -> None:
        """Show the repo in the default browser."""
        webbrowser.open(self.commit_url)

    def exists(self) -> bool:
        if self.repo_type != RepoType.GITHUB:
            repo = self.repo  # git repo
            try:
                repo.commit(self.commit)
                if repo.head.commit.hexsha != self.commit:
                    self.repo.git.checkout(self.commit)
                return True
            except BadName:
                logger.warning(
                    f"Commit {self.commit} does not exist in this repository."
                )
                return False
        else:
            return url_exists(self.commit_url)

    def clone_and_checkout(self) -> None:
        """Clone the repo to the current working directory and checkout a specific commit."""
        logger.debug(f"Cloning {self}")
        repo = Repo.clone_from(self.url, Path(self.name), no_checkout=True)
        repo.git.checkout(self.commit)
        repo.submodule_update(init=True, recursive=True)

    def get_dependencies(
        self, path: Union[str, Path, None] = None
    ) -> Dict[str, "LeanGitRepo"]:
        """Return the dependencies required by the target repo.

        Args:
            path (Union[str, Path, None], optional): Root directory of the repo if it is on the disk.

        Returns:
            Dict[str, :class:`LeanGitRepo`]: A dictionary mapping the name of each
            dependency to its :class:`LeanGitRepo` object.
        """
        logger.debug(f"Querying the dependencies of {self}")

        toolchain = (
            self.get_config("lean-toolchain")
            if path is None
            else {"content": (Path(path) / "lean-toolchain").open().read()}
        )
        commit = get_lean4_commit_from_config(toolchain)
        deps = {"lean4": LeanGitRepo(LEAN4_URL, commit)}

        try:
            lake_manifest = (
                self.get_config("lake-manifest.json", num_retries=0)
                if path is None
                else json.load((Path(path) / "lake-manifest.json").open())
            )
            for pkg in lake_manifest["packages"]:
                deps[pkg["name"]] = LeanGitRepo(pkg["url"], pkg["rev"])
        except Exception:
            for name, repo in self._parse_lakefile_dependencies(path):
                if name not in deps:
                    deps[name] = repo
                for dd_name, dd_repo in repo.get_dependencies().items():
                    deps[dd_name] = dd_repo

        return deps

    def _parse_lakefile_dependencies(
        self, path: Union[str, Path, None]
    ) -> List[Tuple[str, "LeanGitRepo"]]:
        if self.uses_lakefile_lean():
            return self._parse_lakefile_lean_dependencies(path)
        else:
            return self._parse_lakefile_toml_dependencies(path)

    def _parse_lakefile_lean_dependencies(
        self, path: Union[str, Path, None]
    ) -> List[Tuple[str, "LeanGitRepo"]]:
        lakefile = (
            self.get_config("lakefile.lean")["content"]
            if path is None
            else (Path(path) / "lakefile.lean").open().read()
        )

        if _LAKEFILE_LEAN_LOCAL_REQUIREMENT_REGEX.search(lakefile):
            raise ValueError("Local dependencies are not supported.")

        return self._parse_deps(_LAKEFILE_LEAN_GIT_REQUIREMENT_REGEX.finditer(lakefile))

    def _parse_deps(
        self, matches: Union[Iterator[re.Match[str]], Dict[str, str]]
    ) -> List[Tuple[str, "LeanGitRepo"]]:
        deps = []

        for m in matches:
            url = m["url"]  # type: ignore
            if url.endswith(".git"):
                url = url[:-4]
            if url.startswith("git@"):
                url = "https://" + url[4:].replace(":", "/")
            try:
                rev = m["rev"]  # type: ignore
            except KeyError:
                rev = None
            if rev is None:
                commit = get_latest_commit(url)
            elif len(rev) == 40 and _COMMIT_REGEX.fullmatch(rev):
                commit = rev
            else:
                try:
                    commit = _to_commit_hash(url_to_repo(url), rev)
                except ValueError:
                    commit = get_latest_commit(url)
                assert _COMMIT_REGEX.fullmatch(commit)

            deps.append((m["name"], LeanGitRepo(url, commit)))  # type: ignore

        return deps

    def _parse_lakefile_toml_dependencies(
        self, path: Union[str, Path, None]
    ) -> List[Tuple[str, "LeanGitRepo"]]:
        lakefile = (
            self.get_config("lakefile.toml")
            if path is None
            else (Path(path) / "lakefile.toml").open().read()
        )
        # Parsing worked
        if isinstance(lakefile, dict) and "require" in lakefile:
            matches = lakefile["require"]
        else:
            if "content" in lakefile:
                lakefile = lakefile["content"]
            matches = []
            for req in _LAKEFILE_TOML_REQUIREMENT_REGEX.finditer(lakefile):
                match = {}
                for line in req.group().strip().splitlines():
                    key, value = line.split("=")
                    match[key.strip()] = value.strip()
                matches.append(match)
        for match in matches:
            if "path" in match:
                raise ValueError("Local dependencies are not supported.")
            if "git" in match:
                match["url"] = match["git"]
                del match["git"]

        return self._parse_deps(matches)

    def get_license(self) -> Optional[str]:
        """Return the content of the ``LICENSE`` file."""
        if self.repo_type == RepoType.GITHUB:
            assert "github.com" in self.url, f"Unsupported URL: {self.url}"
            url = self.url.replace("github.com", "raw.githubusercontent.com")
            license_url = f"{url}/{self.commit}/LICENSE"
            try:
                return read_url(license_url)
            except urllib.error.HTTPError:  # type: ignore
                return None
        else:
            license_path = Path(self.repo.working_dir) / "LICENSE"
            if license_path.exists():
                return license_path.open("r").read()
            else:
                return None

    def _get_config_url(self, filename: str) -> str:
        assert self.repo_type == RepoType.GITHUB
        assert "github.com" in self.url, f"Unsupported URL: {self.url}"
        url = self.url.replace("github.com", "raw.githubusercontent.com")
        return f"{url}/{self.commit}/{filename}"

    def get_config(self, filename: str, num_retries: int = 2) -> Dict[str, Any]:
        """Return the repo's files."""
        if self.repo_type == RepoType.GITHUB:
            config_url = self._get_config_url(filename)
            content = read_url(config_url, num_retries)
        else:
            working_dir = self.repo.working_dir
            with open(os.path.join(working_dir, filename), "r") as f:
                content = f.read()
        if filename.endswith(".toml"):
            return toml.loads(content)
        elif filename.endswith(".json"):
            return json.loads(content)
        else:
            return {"content": content}

    def uses_lakefile_lean(self) -> bool:
        """Check if the repo uses a ``lakefile.lean``."""
        if self.repo_type == RepoType.GITHUB:
            url = self._get_config_url("lakefile.lean")
            return url_exists(url)
        else:
            lakefile_path = Path(self.repo.working_dir) / "lakefile.lean"
            return lakefile_path.exists()

    def uses_lakefile_toml(self) -> bool:
        """Check if the repo uses a ``lakefile.toml``."""
        if self.repo_type == RepoType.GITHUB:
            url = self._get_config_url("lakefile.toml")
            return url_exists(url)
        else:
            lakefile_path = Path(self.repo.working_dir) / "lakefile.toml"
            return lakefile_path.exists()


@dataclass(frozen=True)
class Theorem:
    """Theorem in Lean.

    Theorems are named constants of type :code:`Prop`. They are typically defined
    using the keywords :code:`theorem` or :code:`lemma`, but it's possible to use other
    keywords such as :code:`def` or :code:`instance`
    """

    repo: LeanGitRepo
    """Lean repo the theorem comes from.
    """

    file_path: Path
    """Lean source file the theorem comes from.
    """

    full_name: str
    """Fully qualified name of the theorem.
    """

    def __post_init__(self) -> None:
        if isinstance(self.file_path, str):
            object.__setattr__(self, "file_path", Path(self.file_path))
        assert (
            self.file_path.suffix == ".lean"
        ), f"File extension must be .lean: {self.file_path}"

    @property
    def uid(self) -> str:
        """Unique identifier of the theorem."""
        return f"{cleanse_string(self.repo.url)}@{cleanse_string(self.repo.commit)}:{cleanse_string(self.file_path.__str__())}:{cleanse_string(self.full_name)}"

    @property
    def uhash(self) -> str:
        """Unique hash of the theorem."""
        return str(hash(self.uid) ** 2)



#!/usr/bin/env python3
"""
Creates a toy LeanProgress dataset showing how to store remaining-step labels.

Each line:
{
    "goal": "... Lean goal ...",
    "prefix": "... optional proof prefix ...",
    "tactic": "... candidate tactic ...",
    "steps_remaining": <non-negative integer>
}
"""

from __future__ import annotations

import argparse
import json
from pathlib import Path
from textwrap import dedent

SAMPLE_DATA = [
    {
        "goal": "n : ℕ ⊢ gcd n n = n",
        "prefix": dedent(
            """
            have : n.gcd n ∣ n := by
              exact gcd_dvd_right n n
            """
        ).strip(),
        "tactic": "simpa [Nat.gcd_comm] using Nat.gcd_self n",
        "steps_remaining": 1,
    },
    {
        "goal": "a b : ℝ, ha : a = b ⊢ b = a",
        "prefix": "",
        "tactic": "exact ha.symm",
        "steps_remaining": 0,
    },
    {
        "goal": "G : Group, g h : G ⊢ g * h = h * g",
        "prefix": "",
        "tactic": "sorry",
        "steps_remaining": 12,
    },
]


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Generate a sample LeanProgress dataset in JSONL format."
    )
    parser.add_argument(
        "--output",
        type=Path,
        default=Path("raid/data/sample_leanprogress_dataset.jsonl"),
        help="Where to save the sample JSONL file.",
    )
    return parser.parse_args()


def main() -> None:
    args = parse_args()
    args.output.parent.mkdir(parents=True, exist_ok=True)
    with args.output.open("w", encoding="utf-8") as f:
        for row in SAMPLE_DATA:
            f.write(json.dumps(row, ensure_ascii=False) + "\n")
    print(f"Wrote {len(SAMPLE_DATA)} LeanProgress examples to {args.output.resolve()}")


if __name__ == "__main__":
    main()



"""
Utility functions for managing repository data, operations, and skipping functionality.
"""

import json
import os
from typing import List, Tuple

from lean_dojo_v2.database.models import Repository
from lean_dojo_v2.utils.constants import DATA_DIR, RAID_DIR


def save_sorted_repos(sorted_repos: List[Repository], file_path: str) -> None:
    """
    Saves the sorted repositories to a file.

    Args:
        sorted_repos: List of Repository objects to save
        file_path: Path to the file where repositories will be saved
    """
    sorted_repo_data = [
        {"url": repo.url, "commit": repo.commit, "name": repo.name}
        for repo in sorted_repos
    ]
    with open(file_path, "w") as f:
        json.dump(sorted_repo_data, f, indent=2)


def load_sorted_repos(file_path: str) -> List[Tuple[str, str, str]]:
    """
    Loads the sorted repositories from a file.

    Args:
        file_path: Path to the file containing repository data

    Returns:
        List of tuples containing (url, commit, name) for each repository
    """
    with open(file_path, "r") as f:
        sorted_repo_data = json.load(f)
    return [(repo["url"], repo["commit"], repo["name"]) for repo in sorted_repo_data]



"""
Benchmark dataset generation for Lean 4 theorem proving.

This module provides functionality to generate training datasets from Lean 4 repositories
by tracing theorem proofs and splitting them into train/validation/test sets.
"""

import json
import os
import random
from collections import defaultdict
from copy import copy
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Union

import networkx as nx
from loguru import logger

from lean_dojo_v2.utils.constants import LEAN4_PACKAGES_DIR, __version__

# Import utility functions from the lean_dojo module
from .lean import LeanGitRepo, get_lean4_version_from_config, is_supported_version
from .trace import trace
from .traced_data import TracedRepo, TracedTheorem

# Set random seed for reproducibility
random.seed(3407)  # https://arxiv.org/abs/2109.08203

# Type aliases for better readability
SPLIT_NAME = str  # train/val/test
SPLIT = Dict[SPLIT_NAME, List[TracedTheorem]]
SPLIT_STRATEGY = str


# =============================================================================
# DATA SPLITTING FUNCTIONS
# =============================================================================


def _split_sequentially(
    traced_theorems: List[TracedTheorem], num_val: int, num_test: int
) -> SPLIT:
    """Split theorems sequentially into train/val/test sets."""
    num_theorems = len(traced_theorems)
    num_train = num_theorems - num_val - num_test

    return {
        "train": traced_theorems[:num_train],
        "val": traced_theorems[num_train : num_train + num_val],
        "test": traced_theorems[num_train + num_val :],
    }


def split_randomly(
    traced_theorems: List[TracedTheorem], num_val: int, num_test: int
) -> SPLIT:
    """Split theorems randomly into train/val/test sets."""
    logger.info("Splitting theorems randomly")
    traced_theorems = copy(traced_theorems)
    random.shuffle(traced_theorems)
    return _split_sequentially(traced_theorems, num_val, num_test)


def split_by_premise(
    traced_theorems: List[TracedTheorem], num_val: int, num_test: int
) -> SPLIT:
    """
    Split theorems so that val/test proofs rely on novel premises not in train.

    This ensures that validation and test sets contain theorems that use premises
    not seen during training, making the evaluation more realistic.
    """
    logger.info("Splitting theorems by novel premises")

    num_val_test = num_val + num_test
    theorems_val_test = set()

    # Map each premise to theorems that use it
    theorems_by_premises = defaultdict(list)
    for theorem in traced_theorems:
        for premise in theorem.get_premise_full_names():
            theorems_by_premises[premise].append(theorem)

    # Sort premises by usage (ascending) to prioritize rare premises
    premises_sorted = sorted(theorems_by_premises.items(), key=lambda x: len(x[1]))

    # Assign theorems to val/test based on novel premises
    for _, theorems in premises_sorted:
        if len(theorems_val_test) < num_val_test:
            theorems_val_test.update(theorems)

    # Remaining theorems go to training
    theorems_train = [t for t in traced_theorems if t not in theorems_val_test]
    theorems_val_test = list(theorems_val_test)
    random.shuffle(theorems_val_test)

    return {
        "train": theorems_train,
        "val": theorems_val_test[:num_val],
        "test": theorems_val_test[num_val:],
    }


def split_data(
    traced_repo: TracedRepo, num_val_pct: float = 0.2, num_test_pct: float = 0.2
) -> Dict[SPLIT_STRATEGY, SPLIT]:
    """
        Split traced theorems into training, validation, and test sets.
    c
        Args:
            traced_repo: Repository containing theorems to split
            num_val_pct: Percentage of theorems for validation (default: 0.2)
            num_test_pct: Percentage of theorems for testing (default: 0.2)

        Returns:
            Dictionary mapping split strategies to their corresponding splits
    """
    # Exclude Lean 4 repository theorems
    traced_theorems = [
        thm for thm in traced_repo.get_traced_theorems() if not thm.repo.is_lean4
    ]

    num_theorems = len(traced_theorems)
    num_val = int(num_theorems * num_val_pct)
    num_test = int(num_theorems * num_test_pct)

    logger.info(
        f"Total theorems: {num_theorems}, " f"Validation: {num_val}, Test: {num_test}"
    )

    return {
        "random": split_randomly(traced_theorems, num_val, num_test),
        "novel_premises": split_by_premise(traced_theorems, num_val, num_test),
    }


# =============================================================================
# EXPORT FUNCTIONS
# =============================================================================


def _get_file_path(traced_repo: TracedRepo, thm: TracedTheorem) -> str:
    """Get the file path for a theorem, handling dependencies."""
    if thm.repo == traced_repo.repo:
        # Theorem belongs to the traced repo itself
        return str(thm.theorem.file_path)
    else:
        # Theorem belongs to a dependency
        for name, dep in traced_repo.dependencies.items():
            if dep == thm.repo:
                return f"{LEAN4_PACKAGES_DIR}/{name}/{thm.theorem.file_path}"
        raise ValueError(f"Unable to find dependency {thm.repo}")


def export_proofs(
    splits: Dict[SPLIT_STRATEGY, SPLIT], dst_path: Path, traced_repo: TracedRepo
) -> int:
    """
    Export theorem proofs to JSON files organized by split strategy.

    Args:
        splits: Dictionary mapping strategies to theorem splits
        dst_path: Destination directory for exported proofs
        traced_repo: Repository containing the theorems

    Returns:
        Total number of theorems exported
    """
    # Collect all unique theorems across all strategies
    # (all strategies contain the same theorems, just split differently)
    unique_theorems = set()
    
    for strategy, split in splits.items():
        strategy_dir = dst_path / strategy
        strategy_dir.mkdir(parents=True)

        for split_name, theorems in split.items():
            data = []
            num_tactics = 0

            for theorem in theorems:
                # Track unique theorems by their identity
                unique_theorems.add(id(theorem))
                
                # Filter out tactics with "no goals" or containing "·"
                tactics = [
                    {
                        "tactic": tactic.tactic,
                        "annotated_tactic": tactic.get_annotated_tactic(),
                        "state_before": tactic.state_before,
                        "state_after": tactic.state_after,
                    }
                    for tactic in theorem.get_traced_tactics()
                    if (tactic.state_before != "no goals" and "·" not in tactic.tactic)
                ]
                num_tactics += len(tactics)

                # Get theorem statement if available
                theorem_statement = None
                if (
                    theorem.has_tactic_proof()
                    and theorem.get_tactic_proof() is not None
                ):
                    theorem_statement = theorem.get_theorem_statement()

                data.append(
                    {
                        "url": traced_repo.repo.url,
                        "commit": traced_repo.repo.commit,
                        "file_path": _get_file_path(traced_repo, theorem),
                        "full_name": theorem.theorem.full_name,
                        "theorem_statement": theorem_statement,
                        "start": list(theorem.start),
                        "end": list(theorem.end),
                        "traced_tactics": tactics,
                    }
                )

            # Save to JSON file
            output_path = strategy_dir / f"{split_name}.json"
            json.dump(data, output_path.open("wt"))

            logger.info(
                f"Saved {len(theorems)} theorems with {num_tactics} tactics "
                f"to {output_path}"
            )

    total_theorems = len(unique_theorems)
    logger.info(f"Total theorems exported: {total_theorems}")
    return total_theorems


def export_premises(traced_repo: TracedRepo, dst_path: Path) -> tuple[int, int]:
    """
    Export premise definitions and traced file information.

    Args:
        traced_repo: Repository containing traced files
        dst_path: Destination directory for exported data

    Returns:
        Tuple of (number of premises, number of traced files)
    """
    # Export premises in topological order
    corpus_path = dst_path / "corpus.jsonl"
    num_premises = 0

    with corpus_path.open("wt") as output:
        graph = traced_repo.traced_files_graph

        if not graph:
            logger.warning("No traced files found in repository")
            pass
        else:
            for node in reversed(list(nx.topological_sort(graph))):
                traced_file = graph.nodes[node]["traced_file"]
                imports = [str(imp) for imp in graph.successors(node)]
                premises = traced_file.get_premise_definitions()
                num_premises += len(premises)

                output.write(
                    json.dumps(
                        {
                            "path": str(traced_file.path),
                            "imports": imports,
                            "premises": premises,
                        }
                    )
                    + "\n"
                )

    logger.info(
        f"Exported {num_premises} premises from "
        f"{len(traced_repo.traced_files)} files to {corpus_path}"
    )

    # Export traced file paths
    traced_files_path = dst_path / "traced_files.jsonl"
    with traced_files_path.open("wt") as output:
        for traced_file in traced_repo.traced_files:
            file_path = traced_file.lean_file.path
            output.write(json.dumps({"traced_file_path": str(file_path)}) + "\n")

    return num_premises, len(traced_repo.traced_files)


def export_metadata(traced_repo: TracedRepo, dst_path: Path, **kwargs) -> None:
    """
    Export repository metadata to JSON file.

    Args:
        traced_repo: Repository containing metadata
        dst_path: Destination directory
        **kwargs: Additional metadata to include
    """
    metadata = dict(kwargs)
    metadata.update(
        {
            "creation_time": str(datetime.now()),
            "from_repo": {
                "url": traced_repo.repo.url,
                "commit": traced_repo.repo.commit,
            },
            "leandojo_version": __version__,
        }
    )

    metadata_path = dst_path / "metadata.json"
    json.dump(metadata, metadata_path.open("wt"))


def export_data(
    traced_repo: TracedRepo,
    splits: Dict[SPLIT_STRATEGY, SPLIT],
    dst_path: Union[str, Path],
    **kwargs,
) -> tuple[int, int, int]:
    """
    Export complete dataset from traced repository.

    Args:
        traced_repo: Repository containing data to export
        splits: Dictionary mapping strategies to theorem splits
        dst_path: Destination directory for exported data
        **kwargs: Additional metadata

    Returns:
        Tuple of (num_premises, num_files_traced, total_theorems)
    """
    if isinstance(dst_path, str):
        dst_path = Path(dst_path)

    # Clear existing directory
    if dst_path.exists():
        import shutil

        shutil.rmtree(dst_path)

    # Export all components
    total_theorems = export_proofs(splits, dst_path, traced_repo)
    num_premises, num_files_traced = export_premises(traced_repo, dst_path)
    export_metadata(traced_repo, dst_path, **kwargs)

    return num_premises, num_files_traced, total_theorems


# =============================================================================
# LEAN ENVIRONMENT SETUP
# =============================================================================


def setup_lean_environment(repo: LeanGitRepo) -> None:
    """
    Configure Lean environment to match repository requirements.

    Args:
        repo: Repository to configure environment for
    """
    # we need to change the toolchain version that LeanAgent uses
    # to match the repo we are currently tracing
    config = repo.get_config("lean-toolchain")
    v = get_lean4_version_from_config(config["content"])
    v = v[1:]  # ignore "v" at beginning

    lean_dir_1 = f"/.elan/toolchains/leanprover--lean4---{v}"
    lean_dir_2 = f"~/.elan/toolchains/leanprover--lean4---{v}"

    if os.path.exists(lean_dir_1):
        lean_dir = lean_dir_1
    else:
        lean_dir = lean_dir_2

    os.environ["LEAN4_PATH"] = lean_dir
    os.environ["PATH"] = f"{lean_dir}/bin:{os.environ.get('PATH', '')}"


# =============================================================================
# MAIN DATASET GENERATION
# =============================================================================


def generate_benchmark(
    repo: LeanGitRepo, dst_dir: str, build_deps: bool
) -> tuple[TracedRepo, int]:
    """
    Generate benchmark dataset from Lean 4 repository.

    This function:
    1. Clones and checks out the specified repository
    2. Configures Lean environment to match repository requirements
    3. Traces the repository to extract theorem proofs
    4. Splits data into train/validation/test sets
    5. Exports all data to the specified directory

    Args:
        url: Git repository URL
        commit: Commit hash to check out
        dst_dir: Destination directory for dataset

    Returns:
        Tuple of (traced_repo, num_premises, num_files_traced, total_theorems)
        Returns (None, 0, 0, 10) if tracing fails
    """
    logger.info(f"Repository: {repo}")

    # Configure Lean environment
    setup_lean_environment(repo)

    # Trace repository
    traced_repo = trace(repo, build_deps=build_deps)
    logger.info("Repository tracing completed successfully")

    # Generate and export dataset
    if os.path.exists(dst_dir):
        import shutil

        shutil.rmtree(dst_dir)

    splits = split_data(traced_repo)
    logger.info("Data splitting completed")

    num_premises, num_files_traced, total_theorems = export_data(
        traced_repo, splits, dst_dir
    )

    logger.info(
        f"Dataset generation complete: "
        f"{total_theorems} theorems, "
        f"{num_premises} premises, "
        f"{num_files_traced} files"
    )

    return traced_repo, total_theorems



"""
Constants and configuration for LeanAgent.
"""

import multiprocessing
import os
import re
import subprocess
from pathlib import Path
from typing import List, Tuple

from dotenv import load_dotenv

# Batch size for training
BATCH_SIZE = 1

# Environment variables
RAID_DIR = os.path.join(os.getcwd(), "raid")
if "GITHUB_ACCESS_TOKEN" not in os.environ:
    raise ValueError("GITHUB_ACCESS_TOKEN environment variable must be set")
GITHUB_ACCESS_TOKEN = os.environ["GITHUB_ACCESS_TOKEN"]

os.environ["RAY_TMPDIR"] = f"/tmp/ray"

# Directory paths
REPO_DIR = "repos_new"
DATA_DIR = "data"
CHECKPOINT_DIR = "checkpoints"
FISHER_DIR = "fisher"  # Optional

# File paths
EVAL_RESULTS_FILE_PATH = f"{RAID_DIR}/eval_results.json"
DB_FILE_NAME = "dynamic_database.json"
PROOF_LOG_FILE_NAME = "proof_log.txt"
ENCOUNTERED_THEOREMS_FILE = f"{RAID_DIR}/encountered_theorems.json"

# List of known repositories to process or skip
# Feel free to remove any repos from this list if you would like to test on them
KNOWN_REPOSITORIES: List[str] = [
    "leanprover-community/mathlib4",  # ReProver is trained on this
    "leanprover-community/batteries",  # functional programming instead of math
    "leanprover-community/aesop",
    "leanprover/lean4",
    "leanprover-community/mathlib",  # Mathlib3 version
    "leanprover-community/mathlib3",
    "leanprover/std4",  # moved to batteries
    "leanprover-community/duper",  # functional programming instead of math
    "leanprover/lake",
    "openai/lean-gym",
    "leanprover-community/lean4-metaprogramming-book",
    "kmill/lean4-raytracer",  # no theorems
    "argumentcomputer/yatima",  # trace problems
    "ImperialCollegeLondon/formalising-mathematics-2024",  # trace problems
    "leanprover-community/ProofWidgets4",  # trace problems
    "leanprover/verso",  # trace problems
    "leanprover-community/NNG4",  # trace problems
    "ufmg-smite/lean-smt",  # fails to trace due to windows-style line endings
    "teorth/symmetric_project",  # no compatible commit
    "cmu-l3/llmlean",  # irrelevant + only 4 theorems
    "PatrickMassot/GlimpseOfLean",  # strange trace problems with _parse_deps
    "avigad/lamr",  # trace problems
    "leanprover-community/quote4",  # no theorems
    "leanprover-community/iris-lean",  # trace problems
    "aripiprazole/rinha",  # incompatible commit
    "leanprover/lean4-cli",  # no theorems
    "leanprover/LeanInk",  # no theorems
    "leanprover-community/lean-auto",
    "leanprover-community/repl",  # no theorems
    "leanprover/doc-gen4",  # no theorems
    "leanprover/SampCert",  # trace problems
    "nomeata/loogle",
    "risc0/risc0-lean4",
    "PatrickMassot/verbose-lean4",  # no theorems
    "tydeu/lean4-alloy",  # no theorems
    "leanprover/leansat",  # deprecated
    "BoltonBailey/formal-snarks-project",  # two theorems
    "dwrensha/lean4-maze",  # two theorems
    "leanprover-community/mathport",  # irrelevant
    "argumentcomputer/LSpec",  # one theorem
    "reaslab/jixia",  # no theorems
    "riccardobrasca/flt3",  # no theorems
    "dwrensha/animate-lean-proofs",  # irrelevant
    "lean-ja/lean-by-example",  # irrelevant
    "NethermindEth/Clear",  # no theorems
    "fgdorais/lean4-parser",  # irrelevant
    "semorrison/lean-training-data",  # irrelevant
    "verse-lab/lean-ssr",  # irrelevant
    "GaloisInc/lean-llvm",  # irrelevant
    "argumentcomputer/Wasm.lean",  # irrelevant
    "NethermindEth/EVMYulLean",  # irrelevant
    "rwbarton/advent-of-lean-4",  # irrelevant
    "leanprover-community/tutorials4",  # irrelevant
    "haruhisa-enomoto/mathlib4-all-tactics",  # irrelevant
    "leanprover/LNSym",
    "leanprover-community/flt-regular",
    "opencompl/lean-mlir-old",
    "rami3l/plfl",
    "HEPLean/HepLean",
    "forked-from-1kasper/ground_zero",
    "verified-optimization/CvxLean",
    "leanprover-community/sphere-eversion",
    "optsuite/optlib",
    "YaelDillies/LeanCamCombi",
    "JamesGallicchio/LeanColls",
    "T-Brick/c0deine",
    "jjdishere/EG",
    "alexkeizer/QpfTypes",
    "fpvandoorn/LeanCourse23",
    "marcusrossel/lean-egg",
    "reilabs/proven-zk",
    "algebraic-dev/soda",
    "leanprover-community/llm",
    "dignissimus/Untangle",
    "argumentcomputer/Megaparsec.lean",
    "emilyriehl/infinity-cosmos",
    "BartoszPiotrowski/lean-premise-selection",
    "djvelleman/HTPILeanPackage",
    "girving/ray",
    "Anderssorby/SDL.lean",
    "pandaman64/lean-regex",
    "brown-cs22/CS22-Lean-2023",
    "hhu-adam/GameSkeleton",
    "FR-vdash-bot/Algorithm",
    "PeterKementzey/graph-library-for-lean4",
    "arthurpaulino/LeanMySQL",
    "arthurpaulino/NumLean",
    "FormalSAT/trestle",
    "nomeata/lean-wf-induct",
    "leanprover/lean4checker",
    "IPDSnelting/tba-2022",
    "digama0/mm-lean4",
    "KislyjKisel/Raylib.lean",
    "algebraic-dev/melp",
    "hhu-adam/Robo",  # same as other tutorials but has lots of sorries
    "hargoniX/socket.lean",
    "kovach/etch",
    "damek/gd-lean",
    "0art0/lean-slides",
    "forked-from-1kasper/lean4-categories",
    "katydid/proofs",
    "alexjbest/leaff",
    "sinhp/Poly",
    "lftcm2023/lftcm2023",  # same as other tutorials but has lots of sorries
    "lean-ja/lean99",
    "leanprover/SHerLOC",
    "Seasawher/mdgen",
    "opencompl/egg-tactic-code",
    "david-christiansen/ssft24",
    "T-Brick/lean2wasm",
    "hargoniX/cpdt-lean",
    "jsm28/AperiodicMonotilesLean",
    "draperlaboratory/ELFSage",
    "rookie-joe/automatic-lean4-compilation",
    "madvorak/fecssk",
    "david-christiansen/bob24",
    "awodey/joyal",
    "BrownCS1951x/fpv2023",  # same as other tutorials but has lots of sorries
    "paulch42/lean-spec",
    "siddhartha-gadgil/MetaExamples",
    "dannypsnl/violet",
    "arthurpaulino/LeanREPL",
    "Kha/do-supplement",
    "joehendrix/lean-sat-checker",
    "ammkrn/timelib",
    "kmill/LeanTeX",
    "leanprover/lean4export",
    "leanprover-community/mathlib3port",
    "brown-cs22/CS22-Lean-2024",  # same as other tutorials but has lots of sorries
    "T-Brick/lean-wasm",
    "crabbo-rave/Soup",
    "argumentcomputer/RustFFI.lean",
    "suhr/tmath",
    "leanprover/leanbv",
    "arthurpaulino/FxyLang",
    "SchrodingerZhu/LeanGccBackend",
    "lecopivo/lean4-karray",
    "ImperialCollegeLondon/M1F-explained",
    "proost-assistant/ProostLean",
    "DavePearce/LeanEVM",
    "algebraic-dev/ash",
    "FormalizedFormalLogic/Arithmetization",
    "cmu-l3/ntp-toolkit",
    "dwrensha/tryAtEachStep",
    "yangky11/lean4-example",
    "T-Brick/DateTime",
    "model-checking/rust-lean-models",
    "MichaelStollBayreuth/EulerProducts",
    "hargoniX/Flame",
    "argumentcomputer/Http.lean",
    "madvorak/vcsp",
    "teorth/newton",
    "apnelson1/Matroid",
    "smorel394/TS1",
    "ianjauslin-rutgers/pythagoras4",
    "mortarsanjaya/IMOSLLean4",
    "dupuisf/BibtexQuery",
    "nomeata/lean-calcify",
    "argumentcomputer/FFaCiL.lean",
    "javra/iit",
    "arthurpaulino/viper",
    "lindy-labs/aegis",
    "PatrickMassot/NNG4",
    "argumentcomputer/YatimaStdLib.lean",
    "fgdorais/lean4-unicode-basic",
    "mhuisi/Uniq",
    "Kha/macro-supplement",
    "chenjulang/rubikcubegroup",
    "arthurpaulino/LeanMusic",
    "argumentcomputer/Ipld.lean",
    "Odomontois/advent2022-lean",
    "kbuzzard/IISc-experiments",  # same as other tutorials but has lots of sorries
    "ykonstant1/InfinitePrimes",
    "alexkassil/natural_number_game_lean4",
    "seewoo5/lean-poly-abc",
    "rah4927/lean-dojo-mew",
    "siddhartha-gadgil/proofs-and-programs-2023",
    "PatrickMassot/lean4-game-server",
    "knowsys/Formale-Systeme-in-LEAN",  # same as other tutorials but has lots of sorries
    "katydid/symbolic-automatic-derivatives",
    "girving/interval",
    "ImperialCollegeLondon/group-theory-experiments",
    "knowsys/CertifyingDatalog",
    "bergmannjg/leanCurl",
    "vasnesterov/HadwigerNelson",
    "FWuermse/lean-postgres",
    "leanprover-community/import-graph",
    "Human-Oriented-ATP/lean-tactics",  # more about tactics than premises
    "paulcadman/lean4-leetcode",
    "argumentcomputer/Lurk.lean",
    "AlexDuchnowski/rubiks-cube",
    "SchrodingerZhu/lean-gccjit",
    "JamesGallicchio/http",
    "jtristan/UnicodeSkipListTableExample",
    "adomani/MA4N1_2023",  # same as other tutorials but has lots of sorries
    "remimimimimi/leansec",
    "hhu-adam/lean-i18n",
    "RemyDegenne/testing-lower-bounds",
    "mariainesdff/LocalClassFieldTheory",
    "AviCraimer/relational-calculus-library-lean4",
    "JLimperg/regensburg-itp-school-2023",
    "jaalonso/Calculemus2",
    "mseri/BET",
    "xubaiw/Reservoir.lean",
    "hargoniX/nest-core",
    "siddhartha-gadgil/Polylean",
    "MichaelStollBayreuth/Weights",
    "sanchace/FRACTRAN",
    "argumentcomputer/Poseidon.lean",
    "madvorak/chomsky",
    "T-Brick/ControlFlow",
    "pa-ba/guarded-lean",
]

# Mark symbols for annotations
MARK_START_SYMBOL = "<a>"
MARK_END_SYMBOL = "</a>"


def remove_marks(s: str) -> str:
    """Remove all :code:`<a>` and :code:`</a>` from ``s``."""
    return s.replace(MARK_START_SYMBOL, "").replace(MARK_END_SYMBOL, "")


load_dotenv()

__version__ = "4.20.0"

# Cache and temporary directories
CACHE_DIR = (
    Path(os.environ["CACHE_DIR"])
    if "CACHE_DIR" in os.environ
    else Path.home() / ".cache/lean_dojo"
).absolute()
"""Cache directory for storing traced repos (see :ref:`caching`).
"""

REMOTE_CACHE_URL = "https://dl.fbaipublicfiles.com/lean-dojo"
"""URL of the remote cache (see :ref:`caching`)."""

DISABLE_REMOTE_CACHE = "DISABLE_REMOTE_CACHE" in os.environ
"""Whether to disable remote caching (see :ref:`caching`) and build all repos locally.
"""

TMP_DIR = Path(os.environ["TMP_DIR"]).absolute() if "TMP_DIR" in os.environ else None
"""Temporary directory used by LeanDojo for storing intermediate files
"""
if TMP_DIR is not None:
    TMP_DIR.mkdir(parents=True, exist_ok=True)


# Process and worker configuration
MAX_NUM_PROCS = 32

NUM_PROCS = int(os.getenv("NUM_PROCS", min(multiprocessing.cpu_count(), MAX_NUM_PROCS)))
"""Number of processes to use
"""

NUM_WORKERS = NUM_PROCS - 1

# Lean 4 configuration
LEAN4_URL = "https://github.com/leanprover/lean4"
"""The URL of the Lean 4 repo."""

LEAN4_PACKAGES_DIR = Path(".lake/packages")
"""The directory where Lean 4 dependencies are stored (since v4.3.0-rc2)."""

LOAD_USED_PACKAGES_ONLY = "LOAD_USED_PACKAGES_ONLY" in os.environ
"""Only load depdendency files that are actually used by the target repo."""

LEAN4_BUILD_DIR = Path(".lake/build")

# Tactic execution limits
TACTIC_CPU_LIMIT = int(os.getenv("TACTIC_CPU_LIMIT", 1))
"""Number of CPUs for executing tactics when interacting with Lean.
"""

TACTIC_MEMORY_LIMIT = os.getenv("TACTIC_MEMORY_LIMIT", "32g")
"""Maximum memory when interacting with Lean.
"""

assert re.fullmatch(r"\d+g", TACTIC_MEMORY_LIMIT)


def check_git_version(min_version: Tuple[int, int, int]) -> None:
    """Check the version of Git installed on the system."""
    res = subprocess.run("git --version", shell=True, capture_output=True, check=True)
    output = res.stdout.decode().strip()
    error = res.stderr.decode()
    assert error == "", error
    m = re.search(r"git version (\d+\.\d+\.\d+)", output)
    assert m, f"Could not parse Git version from: {output}"
    # Convert version number string to tuple of integers
    version = tuple(int(_) for _ in m.group(1).split("."))
    version_str = ".".join(str(_) for _ in version)
    min_version_str = ".".join(str(_) for _ in min_version)
    assert (
        version >= min_version
    ), f"Git version {version_str} is too old. Please upgrade to at least {min_version_str}."


check_git_version((2, 25, 0))



"""DeepSeek agent for Lean theorem proving."""

from typing import Optional

from pantograph.expr import GoalState, Tactic

from lean_dojo_v2.database.models.theorems import Theorem
from lean_dojo_v2.external_api.python.external_models import HFTacticGenerator
from lean_dojo_v2.prover.base_prover import BaseProver


class ExternalProver(BaseProver):
    """DeepSeek-based agent for Lean theorem proving.

    This agent uses the DeepSeek-Prover-V2 model to generate tactics
    for theorem proving without retrieval augmentation.
    """

    def __init__(self, model_name: str = "deepseek-ai/DeepSeek-Prover-V2-671B:novita"):
        super().__init__()
        self.tactic_generator = HFTacticGenerator(model_name=model_name)

    def next_tactic(
        self,
        state: GoalState,
        goal_id: int,
    ) -> Optional[Tactic]:
        """Generate the next tactic using DeepSeek model."""
        return self.tactic_generator.generate(str(state))

    def generate_whole_proof(self, theorem: Theorem) -> str:
        self.theorem = theorem
        return self.tactic_generator.generate_whole_proof(str(self.theorem))



#!/usr/bin/env python3
"""
Test script to verify DeepSeek model loading works properly.
This tests the same model loading as used in train_deepseek_simple.py
"""

import logging

import torch
from transformers import AutoModelForCausalLM, AutoTokenizer

logging.basicConfig(
    format="%(asctime)s - %(levelname)s - %(message)s",
    level=logging.INFO,
)
logger = logging.getLogger(__name__)


def test_deepseek_1_3b_loading():
    """Test loading deepseek-ai/deepseek-coder-1.3b-instruct"""
    model_name = "deepseek-ai/deepseek-coder-1.3b-instruct"

    logger.info(f"Testing model loading: {model_name}")
    logger.info(f"CUDA available: {torch.cuda.is_available()}")

    # Test tokenizer loading
    logger.info("Loading tokenizer...")
    tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True)
    logger.info(f"✓ Tokenizer loaded successfully")
    logger.info(f"  Vocab size: {len(tokenizer)}")
    logger.info(f"  Pad token: {tokenizer.pad_token}")

    # Test model loading
    logger.info("Loading model...")
    model = AutoModelForCausalLM.from_pretrained(
        model_name, trust_remote_code=True, torch_dtype=torch.float32, device_map="auto"
    )
    logger.info(f"✓ Model loaded successfully")
    logger.info(f"  Device: {model.device}")
    logger.info(f"  Dtype: {model.dtype}")

    # Test inference
    logger.info("Testing inference...")
    test_input = "def hello_world():"
    inputs = tokenizer(test_input, return_tensors="pt").to(model.device)
    outputs = model.generate(**inputs, max_length=50, do_sample=False)
    result = tokenizer.decode(outputs[0], skip_special_tokens=True)
    logger.info(f"✓ Inference test passed")
    logger.info(f"  Input: {test_input}")
    logger.info(f"  Output: {result[:100]}...")


def test_deepseek_7b_loading():
    """Test loading DeepSeek-Prover-V2-7B (used in LeanDojo-v2 examples)"""
    model_name = "deepseek-ai/DeepSeek-Prover-V2-7B"

    logger.info(f"\nTesting model loading: {model_name}")

    # Test tokenizer loading
    logger.info("Loading tokenizer...")
    tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True)
    logger.info(f"✓ Tokenizer loaded successfully")

    # Test model loading (but don't actually load the full model to save memory)
    logger.info("Checking model config...")
    from transformers import AutoConfig

    config = AutoConfig.from_pretrained(model_name, trust_remote_code=True)
    logger.info(f"✓ Model config accessible")
    logger.info(f"  Model type: {config.model_type}")



from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple
from xml.sax.saxutils import escape, unescape

from lxml import etree

from lean_dojo_v2.utils.common import (
    camel_case,
    is_optional_type,
    parse_int_list,
    parse_str_list,
    remove_optional_type,
)

from .lean import LeanFile, Pos


@dataclass(frozen=True)
class Node:
    lean_file: LeanFile
    start: Optional[Pos]
    end: Optional[Pos]
    children: List["Node"] = field(repr=False)

    @classmethod
    def from_data(cls, node_data: Dict[str, Any], lean_file: LeanFile) -> "Node":
        subcls = cls._kind_to_node_type(node_data["kind"])
        return subcls.from_data(node_data, lean_file)

    @classmethod
    def _kind_to_node_type(cls, kind: str) -> type["Node"]:
        prefix = "Lean.Parser."
        if kind.startswith(prefix):
            kind = kind[len(prefix) :]
        cls_name = camel_case(kind.replace(".", "_")) + "Node"
        gbs = globals()
        if cls_name in gbs:
            return gbs[cls_name]  # type: ignore
        else:
            # logger.warning(kind)
            return OtherNode

    @classmethod
    def kind(cls: type) -> str:
        return cls.__name__[:-4].lower()

    def traverse_preorder(
        self,
        callback: Callable[["Node", List["Node"]], Any],
        node_cls: Optional[type],
        parents: List["Node"] = [],
    ) -> None:
        if node_cls is None or isinstance(self, node_cls):
            if callback(self, parents):
                return
        for child in self.children:
            child.traverse_preorder(callback, node_cls, parents + [self])

    def traverse_postorder(
        self,
        callback: Callable[["Node", List[Any]], Any],
    ) -> Any:
        return callback(
            self, [child.traverse_postorder(callback) for child in self.children]
        )

    def to_xml(self, parent: etree.Element) -> None:
        tree = etree.SubElement(parent, self.__class__.__name__)

        for k in self.__dataclass_fields__:
            if k in ("lean_file", "children"):
                continue
            v = getattr(self, k)
            if v is not None:
                v = escape(str(v), entities={'"': "&quot;"})
                tree.set(k, v)

        for child in self.children:
            child.to_xml(tree)

    @classmethod
    def from_xml(cls, tree: etree.Element, lean_file: LeanFile) -> "Node":
        subcls = globals()[tree.tag]
        start = Pos.from_str(tree.attrib["start"]) if "start" in tree.attrib else None
        end = Pos.from_str(tree.attrib["end"]) if "end" in tree.attrib else None
        children = [Node.from_xml(subtree, lean_file) for subtree in tree]
        kwargs: Dict[str, Any] = {}

        for field in subcls.__dataclass_fields__.values():
            if field.name in ("lean_file", "start", "end", "children"):
                continue
            v = tree.attrib.get(field.name, None)
            if v is None:
                kwargs[field.name] = None
                continue

            assert isinstance(v, str)
            v = unescape(v, entities={"&quot;": '"'})
            tp = (
                remove_optional_type(field.type)
                if is_optional_type(field.type)
                else field.type
            )
            if tp is Pos:
                kwargs[field.name] = Pos.from_str(v)
            elif tp is Path:
                kwargs[field.name] = Path(v)
            elif tp is List[int]:
                kwargs[field.name] = parse_int_list(v)
            elif tp is List[str]:
                kwargs[field.name] = parse_str_list(v)
            else:
                kwargs[field.name] = v  # type: ignore

        return subcls(lean_file, start, end, children, **kwargs)  # type: ignore

    def get_closure(self) -> Tuple[Optional[Pos], Optional[Pos]]:
        return self.start, self.end


def _parse_pos(
    info: Dict[str, Any], lean_file: LeanFile
) -> Optional[Tuple[Optional[Pos], Optional[Pos]]]:
    if "synthetic" in info and not info["synthetic"]["canonical"]:
        return None

    if (
        "original" in info
    ):  # | original (leading : Substring) (pos : String.Pos) (trailing : Substring) (endPos : String.Pos)
        start, end = info["original"]["pos"], info["original"]["endPos"]
    else:
        assert (
            "synthetic" in info
        )  # | synthetic (pos : String.Pos) (endPos : String.Pos) (canonical := false)
        start, end = info["synthetic"]["pos"], info["synthetic"]["endPos"]

    start = lean_file.position_to_pos(start)
    end = lean_file.position_to_pos(end)

    return start, end


@dataclass(frozen=True)
class AtomNode(Node):
    leading: str
    trailing: str
    val: str

    @classmethod
    def from_data(
        cls, atom_data: Dict[str, Any], lean_file: LeanFile
    ) -> Optional["AtomNode"]:
        info = atom_data["info"]
        start, end = _parse_pos(info, lean_file)

        if "original" in info:
            leading = info["original"]["leading"]
            trailing = info["original"]["trailing"]
        else:
            assert "synthetic" in info
            leading = info["synthetic"]["leading"]
            trailing = info["synthetic"]["trailing"]

        return cls(lean_file, start, end, [], leading, trailing, atom_data["val"])


@dataclass(frozen=True)
class IdentNode(Node):
    leading: str
    trailing: str
    raw_val: str
    val: str

    full_name: Optional[str] = None
    mod_name: Optional[str] = None
    def_path: Optional[str] = None
    def_start: Optional[Pos] = None
    def_end: Optional[Pos] = None

    @classmethod
    def from_data(
        cls, ident_data: Dict[str, Any], lean_file: LeanFile
    ) -> Optional["IdentNode"]:
        info = ident_data["info"]
        start, end = _parse_pos(info, lean_file)
        assert ident_data["preresolved"] == []

        if "original" in info:
            leading = info["original"]["leading"]
            trailing = info["original"]["trailing"]
        else:
            assert "synthetic" in info
            leading = info["synthetic"]["leading"]
            trailing = info["synthetic"]["trailing"]

        return cls(
            lean_file,
            start,
            end,
            [],
            leading,
            trailing,
            ident_data["rawVal"],
            ident_data["val"],
        )

    @property
    def is_mutual(self) -> bool:
        return not isinstance(self.full_name, str)


def is_leaf(node: Node) -> bool:
    return isinstance(node, AtomNode) or isinstance(node, IdentNode)


@dataclass(frozen=True)
class FileNode(Node):
    @classmethod
    def from_data(cls, data: Dict[str, Any], lean_file: LeanFile) -> "FileNode":
        children = []

        def _get_closure(node: Node, child_spans: List[Tuple[Pos, Pos]]):
            if len(child_spans) == 0:
                return node.start, node.end

            child_starts = [s for s, _ in child_spans if s is not None]
            if len(child_starts) == 0:
                start = None
            else:
                start = min(child_starts)

            child_ends = [e for _, e in child_spans if e is not None]
            if len(child_ends) == 0:
                end = None
            else:
                end = max(child_ends)

            if node.start is None:
                object.__setattr__(node, "start", start)
            else:
                assert node.start == start

            if node.end is None:
                object.__setattr__(node, "end", end)
            else:
                assert node.end == end

            return start, end

        for i, d in enumerate(data["commandASTs"]):
            node_data = d["node"]
            if i == 0:
                assert node_data["kind"] == "Lean.Parser.Module.header"
            node = Node.from_data(node_data, lean_file)
            node.traverse_postorder(_get_closure)
            children.append(node)

        return cls(lean_file, lean_file.start_pos, lean_file.end_pos, children)


def _parse_children(node_data: Dict[str, Any], lean_file: LeanFile) -> List[Node]:
    children = []

    for d in node_data["args"]:
        if (
            "node" in d
        ):  # | node   (info : SourceInfo) (kind : SyntaxNodeKind) (args : Array Syntax) : Syntax
            node = Node.from_data(d["node"], lean_file)
        elif "atom" in d:  # | atom   (info : SourceInfo) (val : String) : Syntax
            node = AtomNode.from_data(d["atom"], lean_file)
        elif (
            "ident" in d
        ):  # | ident  (info : SourceInfo) (rawVal : Substring) (val : Name) (preresolved : List Syntax.Preresolved) : Syntax
            node = IdentNode.from_data(d["ident"], lean_file)
        else:
            raise ValueError(d)

        if node is not None:
            children.append(node)

    return children


@dataclass(frozen=True)
class TermAttrkindNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "TermAttrkindNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class TermAttrkindAntiquotNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "TermAttrkindAntiquotNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class IdentAntiquotNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "IdentAntiquotNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)

    def get_ident(self) -> str:
        return "".join(gc.val for gc in self.children if is_leaf(gc))


@dataclass(frozen=True)
class LeanElabCommandCommandIrreducibleDefNode(Node):
    name: Optional[str]
    full_name: Optional[str] = None

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "LeanElabCommandCommandIrreducibleDefNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        if isinstance(children[0], CommandDeclmodifiersAntiquotNode):
            name = None
        else:
            assert isinstance(children[0], CommandDeclmodifiersNode)
            assert (
                isinstance(children[1], AtomNode)
                and children[1].val == "irreducible_def"
            )
            declid_node = children[2]
            assert isinstance(declid_node, CommandDeclidNode)
            ident_node = declid_node.children[0]
            assert isinstance(ident_node, IdentNode)
            name = ident_node.val

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class GroupNode(Node):
    @classmethod
    def from_data(cls, node_data: Dict[str, Any], lean_file: LeanFile) -> "GroupNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class MathlibTacticLemmaNode(Node):
    name: str
    full_name: Optional[str] = None
    _is_private_decl: Optional[bool] = (
        False  # `_is_private` doesn't play well with lxml.
    )

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "MathlibTacticLemmaNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert isinstance(children[0], CommandDeclmodifiersNode)
        assert isinstance(children[1], GroupNode)
        assert (
            isinstance(children[1].children[0], AtomNode)
            and children[1].children[0].val == "lemma"
        )
        declid_node = children[1].children[1]
        assert isinstance(declid_node, CommandDeclidNode)
        ident_node = declid_node.children[0]
        assert isinstance(ident_node, IdentNode)
        name = ident_node.val

        return cls(lean_file, start, end, children, name)

    def is_private(self) -> bool:
        return self._is_private_decl

    def get_proof_node(self) -> Node:
        decl_val_node = self.children[1].children[3]
        if isinstance(
            decl_val_node, (CommandDeclvalsimpleNode, CommandWherestructinstNode)
        ):
            return decl_val_node.children[1]
        else:
            return decl_val_node

    def has_tactic_proof(self) -> bool:
        node = self.get_proof_node()
        return isinstance(node, TermBytacticNode)

    @property
    def is_mutual(self) -> bool:
        return not isinstance(self.name, str)


@dataclass(frozen=True)
class LemmaNode(Node):
    name: str
    full_name: Optional[str] = None
    _is_private_decl: Optional[bool] = (
        False  # `_is_private` doesn't play well with lxml.
    )

    @classmethod
    def from_data(cls, node_data: Dict[str, Any], lean_file: LeanFile) -> "LemmaNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        if isinstance(children[0], CommandDeclmodifiersAntiquotNode):
            name = None
        else:
            assert isinstance(children[0], CommandDeclmodifiersNode)
            assert isinstance(children[1], GroupNode)
            assert (
                isinstance(children[1].children[0], AtomNode)
                and children[1].children[0].val == "lemma"
            )
            declid_node = children[1].children[1]
            assert isinstance(declid_node, CommandDeclidNode)
            ident_node = declid_node.children[0]
            assert isinstance(ident_node, IdentNode)
            name = ident_node.val

        return cls(lean_file, start, end, children, name)

    def is_private(self) -> bool:
        return self._is_private_decl

    def get_proof_node(self) -> Node:
        decl_val_node = self.children[1].children[3]
        if isinstance(
            decl_val_node,
            (
                CommandDeclvalsimpleNode,
                CommandWherestructinstNode,
            ),
        ):
            return decl_val_node.children[1]
        else:
            return decl_val_node

    def has_tactic_proof(self) -> bool:
        node = self.get_proof_node()
        return isinstance(node, TermBytacticNode)

    @property
    def is_mutual(self) -> bool:
        return not isinstance(self.name, str)


@dataclass(frozen=True)
class CommandDeclarationNode(Node):
    name: str
    full_name: Optional[str] = None

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandDeclarationNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        if isinstance(children[0], CommandDeclmodifiersAntiquotNode):
            name = None
        else:
            assert isinstance(children[0], CommandDeclmodifiersNode)
            assert isinstance(
                children[1],
                (
                    CommandDefNode,
                    CommandDefinitionNode,
                    CommandTheoremNode,
                    CommandInductiveNode,
                    CommandClassinductiveNode,
                    CommandStructureNode,
                    CommandInstanceNode,
                    CommandAbbrevNode,
                    CommandOpaqueNode,
                    CommandAxiomNode,
                    CommandExampleNode,
                ),
            )
            name = children[1].name

            if children[0].is_private():
                for child in children:
                    if isinstance(child, CommandTheoremNode):
                        object.__setattr__(child, "_is_private_decl", True)

        return cls(lean_file, start, end, children, name)

    @property
    def is_theorem(self) -> bool:
        return isinstance(self.children[1], CommandTheoremNode)

    def get_theorem_node(self) -> "CommandTheoremNode":
        assert self.is_theorem
        return self.children[1]

    @property
    def is_example(self) -> bool:
        return isinstance(self.children[1], CommandExampleNode)


@dataclass(frozen=True)
class CommandDeclmodifiersAntiquotNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandDeclmodifiersAntiquotNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class CommandDeclmodifiersNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandDeclmodifiersNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)

    def is_private(self) -> bool:
        result = False

        def _callback(node: CommandPrivateNode, _) -> bool:
            nonlocal result
            result = True
            return True

        self.traverse_preorder(_callback, CommandPrivateNode)
        return result


@dataclass(frozen=True)
class CommandPrivateNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandPrivateNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class CommandOpenNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandOpenNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class CommandOpenonlyNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandOpenonlyNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class NullNode(Node):
    @classmethod
    def from_data(cls, node_data: Dict[str, Any], lean_file: LeanFile) -> "NullNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class CommandStructuretkNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandStructuretkNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        assert isinstance(children[0], AtomNode) and children[0].val == "structure"
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class CommandClasstkNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandClasstkNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        assert isinstance(children[0], AtomNode) and children[0].val == "class"
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class CommandStructureNode(Node):
    name: str

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandStructureNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert isinstance(children[0], (CommandStructuretkNode, CommandClasstkNode))
        if isinstance(children[1], CommandDeclidAntiquotNode):
            name = None
        else:
            assert isinstance(children[1], CommandDeclidNode)
            decl_id_node = children[1]
            ident_node = decl_id_node.children[0]

            assert isinstance(ident_node, IdentNode)
            name = ident_node.val

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class CommandInductiveNode(Node):
    name: Optional[str]

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandInductiveNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert isinstance(children[0], AtomNode) and children[0].val == "inductive"

        if isinstance(children[1], CommandDeclidAntiquotNode):
            name = None
        else:
            assert isinstance(children[1], CommandDeclidNode)
            decl_id_node = children[1]
            ident_node = decl_id_node.children[0]
            assert isinstance(ident_node, IdentNode)
            name = ident_node.val

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class CommandClassinductiveNode(Node):
    name: str

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandClassinductiveNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert (
            isinstance(children[0].children[0], AtomNode)
            and children[0].children[0].val == "class"
        )
        assert (
            isinstance(children[0].children[1], AtomNode)
            and children[0].children[1].val == "inductive"
        )

        if isinstance(children[1], CommandDeclidAntiquotNode):
            name = None
        else:
            assert isinstance(children[1], CommandDeclidNode)
            decl_id_node = children[1]
            ident_node = decl_id_node.children[0]
            assert isinstance(ident_node, IdentNode)
            name = ident_node.val

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class TermHoleNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "TermHoleNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        assert len(children) == 1 and isinstance(
            children[0],
            (
                AtomNode,
                TokenAntiquotNode,
            ),
        )
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class LeanBinderidentNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "LeanBinderidentNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        assert len(children) == 1 and isinstance(
            children[0],
            (
                TermHoleNode,
                IdentNode,
                IdentAntiquotNode,
            ),
        )
        return cls(lean_file, start, end, children)

    def get_ident(self) -> Optional[str]:
        if isinstance(self.children[0], TermHoleNode):
            return None
        else:
            assert isinstance(self.children[0], IdentNode)
            return self.children[0].val


@dataclass(frozen=True)
class LeanBinderidentAntiquotNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "LeanBinderidentAntiquotNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)

    def get_ident(self) -> Optional[str]:
        return None


@dataclass(frozen=True)
class StdTacticAliasAliasNode(Node):
    name: str
    full_name: Optional[str] = None

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "StdTacticAliasAliasNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert isinstance(children[0], CommandDeclmodifiersNode)
        assert isinstance(children[1], AtomNode) and children[1].val == "alias"
        if isinstance(children[2], IdentAntiquotNode):
            name = None
        else:
            ident_node = children[2]
            assert isinstance(ident_node, IdentNode)
            name = ident_node.val
        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class StdTacticAliasAliaslrNode(Node):
    name: List[str]
    full_name: Optional[List[str]] = None

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "StdTacticAliasAliaslrNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert isinstance(children[0], CommandDeclmodifiersNode)
        assert isinstance(children[1], AtomNode) and children[1].val == "alias"
        assert isinstance(children[2], AtomNode) and children[2].val == "⟨"
        assert isinstance(children[4], AtomNode) and children[4].val == ","
        assert isinstance(children[6], AtomNode) and children[6].val == "⟩"

        name = []
        assert isinstance(
            children[3], (LeanBinderidentNode, LeanBinderidentAntiquotNode)
        )
        name.append(children[3].get_ident())
        assert isinstance(
            children[5], (LeanBinderidentNode, LeanBinderidentAntiquotNode)
        )
        name.append(children[5].get_ident())
        name = [n for n in name if n is not None]

        return cls(lean_file, start, end, children, name)

    @property
    def is_mutual(self) -> bool:
        return True


@dataclass(frozen=True)
class CommandAbbrevNode(Node):
    name: str

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandAbbrevNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert isinstance(children[0], AtomNode) and children[0].val == "abbrev"
        declid_node = children[1]
        if isinstance(declid_node, CommandDeclidAntiquotNode):
            name = None
        else:
            assert isinstance(declid_node, CommandDeclidNode)
            ident_node = declid_node.children[0]
            assert isinstance(ident_node, IdentNode)
            name = ident_node.val

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class CommandOpaqueNode(Node):
    name: str

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandOpaqueNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert isinstance(children[0], AtomNode) and children[0].val == "opaque"
        declid_node = children[1]
        if isinstance(declid_node, CommandDeclidAntiquotNode):
            name = None
        else:
            assert isinstance(declid_node, CommandDeclidNode)
            ident_node = declid_node.children[0]
            assert isinstance(ident_node, IdentNode)
            name = ident_node.val

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class CommandAxiomNode(Node):
    name: str

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandAxiomNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert isinstance(children[0], AtomNode) and children[0].val == "axiom"
        declid_node = children[1]
        if isinstance(declid_node, CommandDeclidAntiquotNode):
            name = None
        else:
            assert isinstance(declid_node, CommandDeclidNode)
            ident_node = declid_node.children[0]
            assert isinstance(ident_node, IdentNode)
            name = ident_node.val

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class CommandExampleNode(Node):
    name: str

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandExampleNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        assert isinstance(children[0], AtomNode) and children[0].val == "example"
        name = None
        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class CommandInstanceNode(Node):
    name: str

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandInstanceNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        if isinstance(children[0], TermAttrkindAntiquotNode):
            name = None
        else:
            assert isinstance(children[0], TermAttrkindNode)
            assert isinstance(children[1], AtomNode) and children[1].val == "instance"
            if children[3].children != []:
                declid_node = children[3].children[0]
                if isinstance(declid_node, CommandDeclidNode):
                    ident_node = declid_node.children[0]
                    if isinstance(ident_node, IdentNode):
                        name = ident_node.val
                    else:
                        assert isinstance(ident_node, IdentAntiquotNode)
                        name = ident_node.get_ident()
                else:
                    name = None
            else:
                name = None
        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class CommandDefNode(Node):
    name: str

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandDefNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        if isinstance(children[0], TokenAntiquotNode) or isinstance(
            children[1], CommandDeclidAntiquotNode
        ):
            name = None
        else:
            assert isinstance(children[0], AtomNode) and children[0].val == "def"
            assert isinstance(children[1], CommandDeclidNode)
            decl_id_node = children[1]
            ident_node = decl_id_node.children[0]

            if isinstance(ident_node, IdentNode):
                name = ident_node.val
            else:
                assert isinstance(ident_node, IdentAntiquotNode)
                name = ident_node.get_ident()

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class CommandDefinitionNode(Node):
    name: str

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandDefinitionNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        if isinstance(children[0], TokenAntiquotNode) or isinstance(
            children[1], CommandDeclidAntiquotNode
        ):
            name = None
        else:
            assert isinstance(children[0], AtomNode) and children[0].val == "def"
            assert isinstance(children[1], CommandDeclidNode)
            decl_id_node = children[1]
            ident_node = decl_id_node.children[0]

            if isinstance(ident_node, IdentNode):
                name = ident_node.val
            else:
                assert isinstance(ident_node, IdentAntiquotNode)
                name = ident_node.get_ident()

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class CommandDeclidAntiquotNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandDeclidAntiquotNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class CommandDeclidNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandDeclidNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class CommandDeclvalsimpleNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandDeclvalsimpleNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class TokenAntiquotNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "TokenAntiquotNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class CommandDeclvaleqnsNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandDeclvaleqnsNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class CommandWherestructinstNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandWherestructinstNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class CommandDeclsigNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandDeclsigNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class TermExplicitbinderNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "TermExplicitbinderNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class TermTypespecNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "TermTypespecNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class CommandTheoremNode(Node):
    name: str
    full_name: Optional[str] = None
    _is_private_decl: Optional[bool] = (
        False  # `_is_private` doesn't play well with lxml.
    )

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandTheoremNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert isinstance(children[0], AtomNode) and children[0].val == "theorem"

        declid_node = children[1]
        if isinstance(declid_node, CommandDeclidAntiquotNode):
            name = None
        else:
            ident_node = declid_node.children[0]
            if isinstance(ident_node, IdentNode):
                name = ident_node.val
            else:
                assert isinstance(ident_node, IdentAntiquotNode)
                name = ident_node.get_ident()

        if not isinstance(children[1], CommandDeclidAntiquotNode):
            assert isinstance(children[2], CommandDeclsigNode)
            decl_val_node = children[3]
            assert isinstance(
                decl_val_node,
                (
                    CommandDeclvalsimpleNode,
                    CommandDeclvaleqnsNode,
                    CommandWherestructinstNode,
                ),
            )

            if isinstance(decl_val_node, CommandDeclvalsimpleNode):
                assert (
                    isinstance(decl_val_node.children[0], AtomNode)
                    and decl_val_node.children[0].val == ":="
                )
            elif isinstance(decl_val_node, CommandWherestructinstNode):
                assert (
                    isinstance(decl_val_node.children[0], AtomNode)
                    and decl_val_node.children[0].val == "where"
                )

        return cls(lean_file, start, end, children, name)

    def is_private(self) -> bool:
        return self._is_private_decl

    def get_proof_node(self) -> Node:
        decl_val_node = self.children[3]
        if isinstance(
            decl_val_node,
            (
                CommandDeclvalsimpleNode,
                CommandWherestructinstNode,
            ),
        ):
            return decl_val_node.children[1]
        else:
            return decl_val_node

    def has_tactic_proof(self) -> bool:
        node = self.get_proof_node()
        return isinstance(node, TermBytacticNode)

    @property
    def is_mutual(self) -> bool:
        return not isinstance(self.name, str)


@dataclass(frozen=True)
class TermBytacticNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "TermBytacticNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class TacticTacticseq1IndentedAntiquotNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "TacticTacticseq1IndentedAntiquotNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)

    def get_tactic_nodes(
        self, atomic_only: bool = False
    ) -> Generator[Node, None, None]:
        return


@dataclass(frozen=True)
class TacticTacticseqNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "TacticTacticseqNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        assert len(children) == 1 and isinstance(
            children[0],
            (
                TacticTacticseq1IndentedNode,
                TacticTacticseqbracketedNode,
                TacticTacticseq1IndentedAntiquotNode,
            ),
        )
        return cls(lean_file, start, end, children)

    def get_tactic_nodes(
        self, atomic_only: bool = False
    ) -> Generator[Node, None, None]:
        yield from self.children[0].get_tactic_nodes(atomic_only)


@dataclass(frozen=True)
class TacticTacticseq1IndentedNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "TacticTacticseq1IndentedNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        assert len(children) == 1 and isinstance(children[0], NullNode)
        return cls(lean_file, start, end, children)

    def get_tactic_nodes(
        self, atomic_only: bool = False
    ) -> Generator[Node, None, None]:
        for i, tac_node in enumerate(self.children[0].children):
            if i % 2 == 0:
                if not atomic_only or not contains_tactic(tac_node):
                    yield tac_node
            else:
                assert isinstance(tac_node, NullNode) or isinstance(tac_node, AtomNode)


@dataclass(frozen=True)
class TacticTacticseqbracketedNode(Node):
    state_before: Optional[str] = None
    state_after: Optional[str] = None
    tactic: Optional[str] = None

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "TacticTacticseqbracketedNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        assert len(children) == 3
        return cls(lean_file, start, end, children)

    @property
    def tactic_nodes(self) -> List[Node]:
        children = self.children
        if not isinstance(children[0], AtomNode) or children[0].val != "{":
            return []

        assert isinstance(children[1], NullNode)
        assert isinstance(children[2], AtomNode) and children[2].val == "}"
        nodes = []
        for i, tac_node in enumerate(children[1].children):
            if i % 2 == 0:
                nodes.append(tac_node)
            else:
                assert isinstance(tac_node, NullNode) or isinstance(tac_node, AtomNode)
        return nodes

    def get_tactic_nodes(
        self, atomic_only: bool = False
    ) -> Generator[Node, None, None]:
        children = self.children
        if isinstance(children[0], AtomNode) and children[0].val == "{":
            assert isinstance(children[1], NullNode)
            assert isinstance(children[2], AtomNode) and children[2].val == "}"
            for i, tac_node in enumerate(children[1].children):
                if i % 2 == 0:
                    if not atomic_only or not contains_tactic(tac_node):
                        yield tac_node
                else:
                    assert isinstance(tac_node, NullNode) or isinstance(
                        tac_node, AtomNode
                    )


def contains_tactic(node: Node) -> bool:
    result = False

    def _callback(x, _) -> bool:
        if x is not node and isinstance(
            x,
            (
                TacticTacticseq1IndentedNode,
                TacticTacticseqbracketedNode,
            ),
        ):
            nonlocal result
            result = True
            return True

    node.traverse_preorder(_callback, node_cls=None)
    return result


@dataclass(frozen=True)
class ModuleHeaderNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "ModuleHeaderNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class ModulePreludeNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "ModulePreludeNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class ModulePreludeNode(Node):
    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "ModulePreludeNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children)


@dataclass(frozen=True)
class ModuleImportNode(Node):
    module: Optional[str]
    path: Optional[Path] = None

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "ModuleImportNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        module = None
        for child in children:
            if isinstance(child, IdentNode):
                module = child.val

        return cls(lean_file, start, end, children, module)


@dataclass(frozen=True)
class CommandModuledocNode(Node):
    comment: str

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandModuledocNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        assert len(children) == 2 and all(isinstance(_, AtomNode) for _ in children)
        assert children[0].val == "/-!"
        comment = children[1].val
        return cls(lean_file, start, end, children, comment)


@dataclass(frozen=True)
class CommandDoccommentNode(Node):
    comment: str

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandDoccommentNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        assert len(children) == 2 and all(isinstance(_, AtomNode) for _ in children)
        assert children[0].val == "/--"
        comment = children[1].val
        return cls(lean_file, start, end, children, comment)


@dataclass(frozen=True)
class CommandNamespaceNode(Node):
    name: str

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandNamespaceNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert len(children) == 2
        assert isinstance(children[0], AtomNode) and children[0].val == "namespace"
        if isinstance(children[1], IdentNode):
            name = children[1].val
        else:
            name = None

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class CommandSectionNode(Node):
    name: Optional[str]

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandNamespaceNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        if children and isinstance(children[0], OtherNode):
            children = children[1:]

        assert len(children) == 2
        assert isinstance(children[0], AtomNode) and children[0].val == "section"
        assert isinstance(children[1], NullNode)

        if len(children[1].children) == 1 and isinstance(
            children[1].children[0], IdentNode
        ):
            name = children[1].children[0].val
        else:
            name = None

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class CommandNoncomputablesectionNode(Node):
    name: Optional[str]

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandNoncomputablesectionNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert len(children) == 3
        assert isinstance(children[0], AtomNode) and children[0].val == "noncomputable"
        assert isinstance(children[1], AtomNode) and children[1].val == "section"
        assert isinstance(children[2], NullNode)

        if len(children[2].children) == 1 and isinstance(
            children[2].children[0], IdentNode
        ):
            name = children[2].children[0].val
        else:
            name = None

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class CommandEndNode(Node):
    name: Optional[str]

    @classmethod
    def from_data(
        cls, node_data: Dict[str, Any], lean_file: LeanFile
    ) -> "CommandEndNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)

        assert len(children) == 2
        assert isinstance(children[0], AtomNode) and children[0].val == "end"
        assert isinstance(children[1], NullNode)

        if len(children[1].children) == 1 and isinstance(
            children[1].children[0], IdentNode
        ):
            name = children[1].children[0].val
        else:
            name = None

        return cls(lean_file, start, end, children, name)


@dataclass(frozen=True)
class OtherNode(Node):
    kind: str  # type: ignore
    state_before: Optional[str] = None
    state_after: Optional[str] = None
    tactic: Optional[str] = None

    @classmethod
    def from_data(cls, node_data: Dict[str, Any], lean_file: LeanFile) -> "OtherNode":
        assert node_data["info"] == "none"
        start, end = None, None
        children = _parse_children(node_data, lean_file)
        return cls(lean_file, start, end, children, node_data["kind"])


def is_potential_premise_lean4(node: Node) -> bool:
    """Check if ``node`` is a theorem/definition that can be used as a premise."""
    if (isinstance(node, CommandDeclarationNode) and not node.is_example) or isinstance(
        node,
        (
            LemmaNode,
            MathlibTacticLemmaNode,
            LeanElabCommandCommandIrreducibleDefNode,
            StdTacticAliasAliasNode,
            StdTacticAliasAliaslrNode,
        ),
    ):
        return node.name is not None
    else:
        return False


def is_mutual_lean4(node: Node) -> bool:
    return (
        isinstance(node, (IdentNode, CommandTheoremNode, StdTacticAliasAliaslrNode))
        and node.is_mutual
    )



"""
Utility modules for LeanAgent.

This package provides common utilities used across the LeanAgent system.
"""

from .common import (
    _is_deepspeed_checkpoint,
    cpu_checkpointing_enabled,
    load_checkpoint,
    zip_strict,
)
from .constants import *

__all__ = [
    "zip_strict",
    "load_checkpoint",
    "cpu_checkpointing_enabled",
    "_is_deepspeed_checkpoint",
]



"""
Git utilities for LeanAgent.

This module contains functions for cloning repositories, managing branches,
creating pull requests, and finding compatible commits for Lean repositories.
"""

import json
import os
import re
import shutil
import subprocess
from functools import cache
from pathlib import Path
from typing import List, Optional, Tuple

import requests
from loguru import logger

from lean_dojo_v2.lean_dojo import LeanGitRepo
from lean_dojo_v2.utils.common import execute
from lean_dojo_v2.utils.filesystem import working_directory
from lean_dojo_v2.utils.lean import get_lean4_version_from_config, is_supported_version


def clone_repo(repo_url: str) -> Tuple[str, str]:
    """Clone a git repository and return the path to the repository and its sha.

    Args:
        repo_url: The URL of the repository to clone

    Returns:
        Tuple of (repo_path, sha) where repo_path is the local path to the cloned repo
        and sha is the latest commit hash
    """
    repo_name = "/".join(repo_url.split("/")[-2:]).replace(".git", "")
    RAID_DIR = os.environ.get("RAID_DIR")
    if not RAID_DIR:
        RAID_DIR = os.path.join(os.getcwd(), "raid")

    logger.info(f"Cloning {repo_url}")
    logger.info(f"Repo name: {repo_name}")
    repo_name = f"{RAID_DIR}/repos_new/{repo_name}"
    if os.path.exists(repo_name):
        print(f"Deleting existing repository directory: {repo_name}")
        shutil.rmtree(repo_name)
    subprocess.run(["git", "clone", repo_url, repo_name])
    process = subprocess.Popen(["git", "ls-remote", repo_url], stdout=subprocess.PIPE)
    stdout, stderr = process.communicate()
    sha = re.split(r"\t+", stdout.decode("utf-8"))[0]
    return repo_name, sha


def branch_exists(repo_name: str, branch_name: str) -> bool:
    """Check if a branch exists in a git repository.

    Args:
        repo_name: Path to the local repository
        branch_name: Name of the branch to check

    Returns:
        True if the branch exists, False otherwise
    """
    proc = subprocess.run(
        ["git", "-C", repo_name, "branch", "-a"], capture_output=True, text=True
    )
    branches = proc.stdout.split("\n")
    local_branch = branch_name
    remote_branch = f"remote/{branch_name}"
    return any(
        branch.strip().endswith(local_branch) or branch.strip().endswith(remote_branch)
        for branch in branches
    )


def create_or_switch_branch(repo_name: str, branch_name: str, base_branch: str) -> None:
    """Create a branch in a git repository if it doesn't exist, or switch to it if it does.

    Args:
        repo_name: Path to the local repository
        branch_name: Name of the branch to create or switch to
        base_branch: Name of the base branch to merge from
    """
    if not branch_exists(repo_name, branch_name):
        subprocess.run(
            ["git", "-C", repo_name, "checkout", "-b", branch_name], check=True
        )
    else:
        subprocess.run(["git", "-C", repo_name, "checkout", branch_name], check=True)
        subprocess.run(
            [
                "git",
                "-C",
                repo_name,
                "merge",
                base_branch,
                "-m",
                f"Merging {branch_name} into {base_branch}",
            ],
            check=True,
        )


def commit_changes(repo_name: str, commit_message: str) -> bool:
    """Commit changes to a git repository.

    Args:
        repo_name: Path to the local repository
        commit_message: The commit message

    Returns:
        True if changes were committed, False if no changes to commit
    """
    status = subprocess.run(
        ["git", "-C", repo_name, "status", "--porcelain"],
        capture_output=True,
        text=True,
    ).stdout.strip()
    if status == "":
        print("No changes to commit.")
        return False
    subprocess.run(["git", "-C", repo_name, "add", "."], check=True)
    subprocess.run(["git", "-C", repo_name, "commit", "-m", commit_message], check=True)
    return True


def push_changes(repo_name: str, branch_name: str) -> None:
    """Push changes to a git repository.

    Args:
        repo_name: Path to the local repository
        branch_name: Name of the branch to push
    """
    subprocess.run(
        ["git", "-C", repo_name, "push", "-u", "origin", branch_name], check=True
    )


def get_default_branch(repo_full_name: str, personal_access_token: str) -> str:
    """Get the default branch of a repository (default `main`).

    Args:
        repo_full_name: Full name of the repository (e.g., "owner/repo")
        personal_access_token: GitHub personal access token

    Returns:
        The default branch name, or "main" if unable to determine
    """
    url = f"https://api.github.com/repos/{repo_full_name}"
    headers = {
        "Authorization": f"token {personal_access_token}",
        "Accept": "application/vnd.github.v3+json",
    }
    response = requests.get(url, headers=headers)
    if response.status_code == 200:
        return response.json()["default_branch"]
    else:
        logger.info(f"Failed to get default branch for {repo_full_name}")
        return "main"


def create_pull_request(
    repo_full_name: str,
    title: str,
    body: str,
    head_branch: str,
    personal_access_token: str,
) -> str:
    """Create a pull request in a repository.

    Args:
        repo_full_name: Full name of the repository (e.g., "owner/repo")
        title: Title of the pull request
        body: Body of the pull request
        head_branch: Name of the branch to merge
        personal_access_token: GitHub personal access token

    Returns:
        URL of the created pull request, or empty string if creation failed
    """
    base_branch = get_default_branch(repo_full_name, personal_access_token)
    url = f"https://api.github.com/repos/{repo_full_name}/pulls"
    headers = {
        "Authorization": f"token {personal_access_token}",
        "Accept": "application/vnd.github.v3+json",
    }
    data = {"title": title, "body": body, "head": head_branch, "base": base_branch}
    response = requests.post(url, headers=headers, json=data)
    if response.status_code == 201:
        print("Pull request created successfully: " + response.json()["html_url"])
        return response.json()["html_url"]
    else:
        print("Failed to create pull request", response.text)
        return ""


def get_compatible_commit(url: str) -> Tuple[Optional[str], Optional[str]]:
    """Find the most recent commit with a Lean version that LeanAgent supports.

    Args:
        url: The repository URL

    Returns:
        Tuple of (commit_sha, lean_version) or (None, None) if no compatible commit found
    """
    try:
        process = subprocess.Popen(["git", "ls-remote", url], stdout=subprocess.PIPE)
        stdout, stderr = process.communicate()
        latest_commit = re.split(r"\t+", stdout.decode("utf-8"))[0]
        logger.info(f"Latest commit: {latest_commit}")

        new_url = url.replace(".git", "")
        logger.info(f"Creating LeanGitRepo for {new_url}")
        repo = LeanGitRepo(new_url, latest_commit)
        logger.info(f"Getting config for {url}")
        config = repo.get_config("lean-toolchain")
        v = get_lean4_version_from_config(config["content"])
        if is_supported_version(v):
            logger.info(f"Latest commit compatible for url {url}")
            return latest_commit, v

        logger.info(f"Searching for compatible commit for {url}")
        try:
            subprocess.run(
                ["git", "rev-parse", "--is-inside-work-tree"],
                check=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            logger.info("Already in a Git repository")
        except subprocess.CalledProcessError:
            logger.info("Not in a Git repository. Initializing one.")
            subprocess.run(["git", "init"], check=True)

        process = subprocess.Popen(
            ["git", "fetch", "--depth=1000000", url],  # Fetch commits
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        logger.info(f"Fetching commits for {url}")
        _, stderr = process.communicate()
        if process.returncode != 0:
            raise Exception(f"Git fetch command failed: {stderr.decode('utf-8')}")
        logger.info(f"Fetched commits for {url}")
        process = subprocess.Popen(
            ["git", "log", "--format=%H", "FETCH_HEAD"],  # Get list of commits
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        logger.info(f"Getting list of commits for {url}")
        stdout, stderr = process.communicate()
        if process.returncode != 0:
            raise Exception(f"Git log command failed: {stderr.decode('utf-8')}")
        commits = stdout.decode("utf-8").strip().split("\n")
        logger.info(f"Found {len(commits)} commits for {url}")
        for commit in commits:
            new_url = url.replace(".git", "")
            repo = LeanGitRepo(new_url, commit)
            config = repo.get_config("lean-toolchain")
            v = get_lean4_version_from_config(config["content"])
            if is_supported_version(v):
                logger.info(f"Found compatible commit {commit} for {url}")
                return commit, v

        raise Exception("No compatible commit found")

    except Exception as e:
        logger.info(f"Error in get_compatible_commit: {str(e)}")
        return None, None


def find_and_save_compatible_commits(
    repo_info_file: str, lean_git_repos: List[LeanGitRepo]
) -> List[dict]:
    """Finds compatible commits for various repositories.

    Args:
        repo_info_file: Path to save the repository information
        lean_git_repos: List of LeanGitRepo objects

    Returns:
        List of dictionaries containing repository information with compatible commits
    """
    updated_repos = []
    for repo in lean_git_repos:
        url = repo.url
        if not url.endswith(".git"):
            url = url + ".git"

        sha = None
        v = None
        sha, v = get_compatible_commit(url)
        if not sha:
            logger.info(f"Failed to find a compatible commit for {url}")
            continue

        updated_repos.append(
            {"url": url.replace(".git", ""), "commit": sha, "version": v}
        )

    with open(repo_info_file, "w") as f:
        json.dump(updated_repos, f)

    return updated_repos


def search_github_repositories(
    language: str = "Lean",
    num_repos: int = 10,
    known_repositories: List[str] = None,
) -> List[LeanGitRepo]:
    """Search for the given number of repositories on GitHub that have the given language.

    Args:
        language: Programming language to search for
        num_repos: Number of repositories to find
        known_repositories: List of repository names to skip

    Returns:
        List of LeanGitRepo objects for the found repositories
    """
    personal_access_token = os.environ.get("GITHUB_ACCESS_TOKEN")

    headers = {"Authorization": personal_access_token}
    query_params = {
        "q": f"language:{language}",
        "sort": "stars",
        "order": "desc",
        "per_page": 100,
    }

    cloned_count = 0
    page = 1
    lean_git_repos = []

    while cloned_count < num_repos:
        query_params["page"] = page
        response = requests.get(
            "https://api.github.com/search/repositories",
            headers=headers,
            params=query_params,
        )

        if not response.status_code == 200:
            logger.info("Failed to search GitHub", response.status_code)
            break

        repositories = response.json()["items"]
        logger.info(f"Found {len(repositories)} repositories")
        for repo in repositories:
            if cloned_count >= num_repos:
                break

            repo_full_name = repo["full_name"]
            logger.info(f"Processing {repo_full_name}")
            if not repo_full_name not in known_repositories:
                logger.info(f"Skipping {repo_full_name} since it is a known repository")
                continue
            name = None
            try:
                clone_url = repo["clone_url"]
                repo_name, sha = clone_repo(clone_url)
                name = repo_name
                url = clone_url.replace(".git", "")
                lean_git_repo = LeanGitRepo(url, sha)
                lean_git_repos.append(lean_git_repo)
                cloned_count += 1
                logger.info(f"Cloned {repo_full_name}")
            except Exception as e:
                if name:
                    shutil.rmtree(name)
                logger.info(f"Failed to clone {repo_full_name} because of {e}")
        page += 1

        # Check if we've reached the end of the search results
        if len(repositories) < 100:
            break

    logger.info(f"Total repositories processed: {cloned_count}")
    return lean_git_repos


@cache
def get_repo_info(path: Path) -> Tuple[str, str]:
    """Get the URL and commit hash of the Git repo at ``path``.

    Args:
        path (Path): Path to the Git repo.

    Returns:
        Tuple[str, str]: URL and (most recent) hash commit
    """
    with working_directory(path):
        # Get the URL.
        url_msg, _ = execute(f"git remote get-url origin", capture_output=True)
        url = url_msg.strip()
        # Get the commit.
        commit_msg, _ = execute(f"git log -n 1", capture_output=True)
        m = re.search(r"(?<=^commit )[a-z0-9]+", commit_msg)
        assert m is not None
        commit = m.group()

    if url.startswith("git@"):
        assert url.endswith(".git")
        url = url[: -len(".git")].replace(":", "/").replace("git@", "https://")

    return url, commit


def get_version_for_commit(url: str, commit: str) -> str:
    repo = LeanGitRepo(url, commit)
    config = repo.get_config("lean-toolchain")
    v = get_lean4_version_from_config(config["content"])
    if not is_supported_version(v):
        raise Exception(f"Unsupported Lean version {v} for {url} at commit {commit}")

    return v



"""Data module for the tactic generator."""

import json
import os
import pickle
from typing import Any, Dict, List, Optional

import pytorch_lightning as pl
from loguru import logger
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm
from transformers import AutoTokenizer, ByT5Tokenizer

from lean_dojo_v2.lean_agent.common import (
    Batch,
    Corpus,
    Example,
    format_augmented_state,
    format_state,
    format_tactic,
)
from lean_dojo_v2.utils.constants import remove_marks


class GeneratorDataset(Dataset):
    """
    A PyTorch Dataset for loading and processing data for a generator model that produces tactics given proof states.

    This dataset handles loading examples from a JSON file, formatting states and tactics,
    and optionally augmenting states with retrieved premises.

    Attributes:
        corpus (Corpus): The corpus containing the proof data.
        keep_marks (bool): Whether to keep markup in the tactics and states.
        preds (List[Dict[str, Any]]): Predictions for augmenting states with retrieved premises.
        max_inp_seq_len (int): Maximum input sequence length for states.
        max_oup_seq_len (int): Maximum output sequence length for tactics.
        p_drop (float): Probability to drop retrieved premises during training.
        tokenizer (ByT5Tokenizer): Tokenizer for encoding states and tactics.
        is_train (bool): Whether this dataset is used for training.
        data (List[Example]): The loaded and processed examples.
    """

    def __init__(
        self,
        data_path: str,
        corpus: Corpus,
        keep_marks: bool,
        preds: List[Dict[str, Any]],
        max_inp_seq_len: int,
        max_oup_seq_len: int,
        p_drop: float,
        normalize_tactics: bool,
        tokenizer: ByT5Tokenizer,
        is_train: bool,
    ) -> None:
        super().__init__()
        self.corpus = corpus
        self.keep_marks = keep_marks
        self.preds = preds
        self.max_inp_seq_len = max_inp_seq_len
        self.max_oup_seq_len = max_oup_seq_len
        self.p_drop = p_drop
        self.tokenizer = tokenizer
        self.is_train = is_train
        self.data = self._load_data(data_path, normalize_tactics)

    def _load_data(self, data_path: str, normalize_tactics: bool) -> List[Example]:
        data = []
        for thm in tqdm(json.load(open(data_path))):
            for tac in thm["traced_tactics"]:
                if "annotated_tactic" in tac:
                    tactic = format_tactic(*tac["annotated_tactic"], normalize_tactics)
                else:
                    tactic = format_tactic(tac["tactic"], [], normalize_tactics)
                if not self.keep_marks:
                    tactic = remove_marks(tactic)
                data.append(
                    {
                        "url": thm["url"],
                        "commit": thm["commit"],
                        "file_path": thm["file_path"],
                        "full_name": thm["full_name"],
                        "state": format_state(tac["state_before"]),
                        "tactic": tactic,
                    }
                )

        logger.info(f"{len(data)} examples loaded")
        return data

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx: int) -> Example:
        ex = self.data[idx]

        if self.preds is not None:
            file_path = ex["file_path"]
            pred = self.preds[(file_path, ex["full_name"], ex["state"])]
            ex["state"] = format_augmented_state(
                ex["state"],
                pred["retrieved_premises"],
                self.max_inp_seq_len,
                self.p_drop if self.is_train else 0.0,
            )

        if not self.keep_marks:
            ex["state"] = remove_marks(ex["state"])

        return ex

    def collate(self, examples: List[Example]) -> Batch:
        state = [ex["state"] for ex in examples]
        tokenized_state = self.tokenizer(
            state,
            padding="longest",
            max_length=self.max_inp_seq_len,
            truncation=True,
            return_tensors="pt",
        )
        tactic = [ex["tactic"] for ex in examples]
        tokenized_tactic = self.tokenizer(
            tactic,
            padding="longest",
            max_length=self.max_oup_seq_len,
            truncation=True,
            return_tensors="pt",
        )
        tactic_ids = tokenized_tactic.input_ids
        tactic_ids[tactic_ids == self.tokenizer.pad_token_id] = -100

        batch = {}
        batch["state"] = state
        batch["state_ids"] = tokenized_state.input_ids
        batch["state_mask"] = tokenized_state.attention_mask
        batch["tactic"] = tactic
        batch["tactic_ids"] = tactic_ids
        batch["tactic_mask"] = tokenized_tactic.attention_mask

        # Copy other fields.
        for k in examples[0].keys():
            if k not in batch:
                batch[k] = [ex[k] for ex in examples]

        return batch


class GeneratorDataModule(pl.LightningDataModule):
    def __init__(
        self,
        data_path: str,
        keep_marks: bool,
        model_name: str,
        batch_size: int,
        eval_batch_size: int,
        max_inp_seq_len: int,
        max_oup_seq_len: int,
        p_drop: float,
        normalize_tactics: bool,
        num_workers: int,
        corpus_path: Optional[str] = None,
        preds_path: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.data_path = data_path
        if corpus_path is not None:
            self.corpus = Corpus(corpus_path)
        else:
            self.corpus = None
        self.keep_marks = keep_marks
        self.batch_size = batch_size
        self.eval_batch_size = eval_batch_size
        self.max_inp_seq_len = max_inp_seq_len
        self.max_oup_seq_len = max_oup_seq_len
        self.p_drop = p_drop
        self.normalize_tactics = normalize_tactics
        self.num_workers = num_workers
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)

        if preds_path is None:
            logger.info("Without retrieval data")
            self.preds = None
        else:
            logger.info("With retrieval data")
            self.preds = {}
            for pred in pickle.load(open(preds_path, "rb")):
                ctx = pred["context"]
                self.preds[ctx.path, ctx.theorem_full_name, ctx.state] = pred

    def prepare_data(self) -> None:
        pass

    def setup(self, stage: Optional[str] = None) -> None:
        if stage in (None, "fit"):
            self.ds_train = GeneratorDataset(
                os.path.join(self.data_path, "train.json"),
                self.corpus,
                self.keep_marks,
                self.preds,
                self.max_inp_seq_len,
                self.max_oup_seq_len,
                self.p_drop,
                self.normalize_tactics,
                self.tokenizer,
                is_train=True,
            )

        if stage in (None, "fit", "validate"):
            self.ds_val = GeneratorDataset(
                os.path.join(self.data_path, "val.json"),
                self.corpus,
                self.keep_marks,
                self.preds,
                self.max_inp_seq_len,
                self.max_oup_seq_len,
                self.p_drop,
                self.normalize_tactics,
                self.tokenizer,
                is_train=False,
            )

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self.ds_train,
            self.batch_size,
            num_workers=self.num_workers,
            collate_fn=self.ds_train.collate,
            shuffle=True,
            pin_memory=True,
            drop_last=True,
        )

    def val_dataloader(self) -> DataLoader:
        return DataLoader(
            self.ds_val,
            self.eval_batch_size,
            num_workers=self.num_workers,
            collate_fn=self.ds_val.collate,
            shuffle=False,
            pin_memory=True,
            drop_last=False,
        )



from .annotations import AnnotatedTactic, Annotation
from .premises import Premise, PremiseFile
from .repository import Repository
from .theorems import Theorem

__all__ = [
    "Annotation",
    "AnnotatedTactic",
    "Theorem",
    "Premise",
    "PremiseFile",
    "Repository",
]
